from .external_mappings import ExternalMapping, ExternalMappingElement, EmptyMapping
from .interpolation_type import InterpolationType

__all__ = ['ExternalMapping', 'ExternalMappingElement', 'EmptyMapping', 'InterpolationType']



import ctypes
from ctypes import CDLL
import numpy as np
import os

from numba.extending import get_cython_function_address as gcf

libc = CDLL(os.path.abspath("cython.cpython-37m-x86_64-linux-gnu.so"))
solve_qr = libc.__pyx_f_8solve_qr_solve_qr_fun




#addr = gcf("cython", "cython")

A = np.random.rand(3,3)
b = np.random.rand(3)
a = A @ A.T

Q, R = np.linalg.qr(a)

Q = np.ascontiguousarray(Q)
R = np.ascontiguousarray(R)
b = np.ascontiguousarray(b)

solve_qr(ctypes.c_int(1), ctypes.c_int(2), ctypes.c_int(3))



# https://www.semicolonworld.com/question/59387/how-to-convert-pointer-to-c-array-to-python-array


from numerous_solver.compile_decorators import basic_njit as njit
import time
import numpy as np
from scipy import linalg

@njit
def forward_substitution_old(L, b):
    #print(L)
    y = np.zeros(len(b))
    y[0] = b[0] / L[0, 0]
    for i in range(1, len(b)):
        v1 = L[i, 0:i]
        v2 = y[0:i]
        prod = v1 @ v2
        y[i] = (b[i] - prod) / L[i, i]
        #y[i] = (b[i] - np.dot(v1,v2)) / L[i, i]
    return y

#@profile
@njit
def forward_substitution(L, b):
    y = np.zeros(len(b))
    Ldiag = np.diag(L)#.tolist()
    y[0] = b[0] / L[0, 0]

    for i in range(1, len(b)):
        v1 = L[i,:i]
        prod = v1 @ y[:i]
        y[i] = (b[i] - prod) / Ldiag[i]


    return y

@njit
def backward_substitution_old(U, y):
    x = np.zeros(len(y))
    n = len(y) - 1
    x[n] = y[n]/U[n,n]
    for i in range(n-1,-1, -1):
        v1 = U[i, (i+1):(n+1)]
        v2 = x[(i+1):(n+1)]
        prod = v1 @ v2
        x[i] = (y[i] - prod) / U[i, i]

    return x

#@profile
@njit
def backward_substitution(U, y):
    x = np.zeros(len(y))
    n = len(y) - 1
    N = n+1

    U_ravel=U.ravel()
    Udiag = np.diag(U)#U_ravel[diag_idx]
    x[n] = y[n]/Udiag[n]

    for i in range(n-1,-1, -1):
        #v1 = U[i, :]
        v1 = U_ravel[(N+1)*i+1:((i+1)*N)]
        v2 = np.array(x[(i+1):(n+1)])
        prod = v1 @ v2#np.dot(v1, v2)
        x[i] = (y[i] - prod) / Udiag[i]
        #x[i] = (y[i] - np.dot(v1,v2))/ U[i, i]
    return np.array(x)

@njit
def solve_cholesky(a, b):
    L = np.linalg.cholesky(a)#
    #y = linalg.solve_triangular(L, b, lower=True)
    #x = linalg.solve_triangular(L.T, y, lower=False)
    y=forward_substitution(L, b)
    x=backward_substitution(np.ascontiguousarray(L.T), y)
    return x


#@profile
@njit
def solve_triangular(L, b):
    y = forward_substitution(L, b)
    x = backward_substitution_old(np.ascontiguousarray(L.T), y)
    #y = linalg.solve_triangular(L,b, lower=True)
    #x = linalg.solve_triangular(L.T, y, lower=False)

    return x

@njit
def solve_qr(Q, R, b):
    bh = Q.T @ b
    x = backward_substitution_old(R, bh)
    return x

@njit
def solve_LU(L, U, b):
    y = forward_substitution(L, b)
    x = backward_substitution_old(U, y)
    return x


def solve_qr_tot(A, b):
    Q, R = np.linalg.qr(A)
    bh = Q.T @ b
    x = backward_substitution(R, bh)
    return x

def timeit(_func, *args):
    start=time.time()
    val=_func(*args)
    end = time.time()
    dt=end-start
    return val, dt

if __name__ == "__main__":
    # a = np.zeros((N, N))
    N = 1000
    a = np.random.rand(N, N)
    a = np.matmul(a, a.T)

    b = np.random.rand(N)



    """
    print("forward substitution:")

    print("myalgo")
    xf_myalgo = forward_substitution(L, b)
    print(xf_myalgo)
    print("scipy")
    xf_scipy = scipy.linalg.solve_triangular(L, b, lower=True)
    print(xf_scipy)

    print("backwards substitution:")

    print("myalgo")
    xb_myalgo = backward_substitution(L.T.conj(), b)
    print(xb_myalgo)
    print("scipy")
    xb_scipy = scipy.linalg.solve_triangular(L.T, b, lower=False)
    print(xb_scipy)
    
    """
    #solve_triangular(a, b)

    x_scipy,dt_scipy=timeit(np.linalg.solve, a,b)
    x_myalgo,dt_myalgo=timeit(solve_qr_tot, a,b)

    print(np.sum(b-np.dot(a,x_scipy)), dt_scipy)
    print(np.sum(b-np.dot(a,x_myalgo)), dt_myalgo)





import logging
import uuid
from numerous.utils.dict_wrapper import _DictWrapper
from numerous.engine.variables import Variable, VariableDescription, _VariableFactory, SetOfVariables


class VariableNamespaceBase:
    """
    Represents a set of variables.

    """

    def __init__(self, item, tag, is_connector=False, _id=uuid.uuid1()):
        self.items = [item.id]
        self.is_connector = is_connector
        self.item = item
        self.set_variables = []
        self.id = str(_id)
        # self.variable_scope = []
        ## -1 outgoing
        ## 0 no mapping
        ## Currently only used in SetNamespace
        self.mappings = []
        self.tag = tag
        self.path = [tag]
        self.outgoing_mappings = 0
        self.associated_equations = {}
        self.variables = _DictWrapper(self.__dict__, Variable)
        self.variable_scope = [self.variables]
        self.registered = False
        self.part_of_set=False
        ## if it is part of set it will be ignored during assembly.
        # print('creating ns with tag: ',tag)

    def get_path_dot(self):
        return ".".join(self.path)

    def __getitem__(self, y):
        return self.variables[y]

    def __setattr__(self, name, value):
        if isinstance(value, Variable):

            self.outgoing_mappings += 1
            self[name].add_mapping(value)
        else:
            object.__setattr__(self, name, value)

    def create_variable(self, name):
        """
        Creates a variable in the namespaces with given name.

        Parameters
        ----------
        name: string
            Name of a 'Variable'

        """
        self.create_variable_from_desc(VariableDescription(tag=name))

    def create_variable_from_desc(self, variable_description):
        """
        Creates and register a variable from given description.

        Parameters
        ----------
        variable_description: 'VariableDescription'
            variable_description
        on_assign_overload : 'OverloadAction'
            action on assign overload
               """

        name = variable_description.tag
        if variable_description.update:
            self[name].model.write_variables(variable_description.initial_value, self[name].llvm_idx)
        else:
            variable = _VariableFactory._create_from_variable_desc(self, self.item, variable_description)
            self.register_variable(variable)

    def get_flat_variables(self):
        return self.variable_scope

    def get_variable(self, var_tag):
        """
        Get a variable with given description.

        Parameters
        ----------
        var_description: 'VariableDescription'
            variable_description

         Returns
        -------
        variable : 'Variable'
                returns a variable matching description or None

        """
        if var_tag in self.variables.internal_dict.keys():
            return self.variables.internal_dict[var_tag]
        else:
            return None

    def register_variable(self, variable, tag_count=""):
        """
        Registering existing Variable in the namespace.

        Parameters
        ----------
        variable: 'Variable'
            Variable to be registered.

        """
        if (variable.tag + tag_count) not in self.variables.keys():
            self.variables[variable.tag + tag_count] = variable

            variable.path.extend_path(variable.id, self.id, self.tag)
            variable.path.extend_path(self.id, self.item.id, self.item.tag)

        else:
            logging.warning("Variable {0} is already in namespace {1} of item {2}".format(variable.tag,
                                                                                          self.tag, self.item.tag))
            # we overwrite constant < parameters < state
            if self.variables[variable.tag].value < variable.value:
                self.variables[variable.tag] = variable
                variable.extend_path(self.tag)
                variable.extend_path(self.item.tag)

    def update_set_var(self):
        pass

    def clear_equations(self):
        self.associated_equations = {}

    def add_equations(self, list_of_equations, update_bindings=True, create_variables=True, set_equation=False):
        """
        Adding a list of equations to namespace. Each equation in the list is parsed and all
         required variables are created and registered in the namespace.

        Parameters
        ----------
        list_of_equations: list of 'Equation'
            list of equations to be added

        update_bindings: bool
            if True creates and register a binding variables in all bindings associated with item
             that namespace is created in.


        """
        if update_bindings and self.is_connector:
            self.item.update_bindings(list_of_equations, self.tag)

        for eq in list_of_equations:
            if create_variables:
                any(self.create_variable_from_desc(variable_description)
                    for variable_description in eq.variables_descriptions)
            eq.set_equation = set_equation
            self.associated_equations.update({eq.tag: eq})


class VariableNamespace(VariableNamespaceBase):
    pass


class SetNamespace(VariableNamespace):
    def __init__(self, item, tag, item_indcs):
        super().__init__(item, tag)
        self.tag = tag
        self.items_id = item_indcs
        self.items = []
        self.len_items = len(self.items)
        self.set_variables = {}

    def add_item(self, item, ns):
        self.items.append(item)
        item_ix = self.items_id.index(ns.item.id)
        for variable in ns.variables:
            if variable.tag not in self.set_variables.keys():
                self.set_variables[variable.tag] = SetOfVariables(variable.tag, self.item.tag, ns.tag)
            self.set_variables[variable.tag].add_variable(variable)
            variable.set_var_ix = item_ix

    def get_flat_variables(self):
        flat_variables = []
        for set_of_variables in self.set_variables.values():
            flat_variables.append(set_of_variables)
        return  flat_variables


class _BindingVariable(Variable):
    def __init__(self, variable):
        super().__init__(variable.detailed_description)
        self.__dict__ = variable.__dict__


class _ShadowVariableNamespace(VariableNamespaceBase):
    def __init__(self, item, tag, binding, is_connector=False, _id=uuid.uuid1()):
        super().__init__(item, tag, is_connector, _id)
        self.binding = binding

    def register_variable(self, variable):
        if variable.tag not in self.variables:
            self.variables[variable.tag] = _BindingVariable(variable)
            self.variables[variable.tag].path.extend_path(variable.id, self.id, self.tag)



import ast

from numerous.engine.model.graph_representation import MappingsGraph, Graph
from numerous.engine.model.graph_representation.utils import EdgeType
from numerous.engine.model.utils import wrap_function


def get_eq_prefix():
    import random, string
    return ''.join(random.choice(string.ascii_letters) for i in range(4)) + '_'


def node_to_ast(n: int, g: MappingsGraph, var_def, ctxread=False, read=True):
    nk = g.key_map[n]
    try:
        if (na := g.get(n, 'ast_type')) == ast.Attribute:

            return var_def(nk, ctxread, read)

        elif na == ast.Name:
            return var_def(nk, ctxread, read)

        elif na == ast.Num:
            return ast.Call(args=[ast.Num(value=g.get(n, 'value'),
                                          lineno=0, col_offset=0)], func=ast.Name(id='float64', lineno=0, col_offset=0,
                                                                                  ctx=ast.Load()),
                            keywords=[], lineno=0, col_offset=0)

        elif na == ast.BinOp:

            left_node = g.get_edges_for_node_filter(end_node=n, attr='e_type', val=EdgeType.LEFT)[1][0][0]

            left_ast = node_to_ast(left_node, g, var_def, ctxread=ctxread)

            right_node = g.get_edges_for_node_filter(end_node=n, attr='e_type', val=EdgeType.RIGHT)[1][0][0]

            right_ast = node_to_ast(right_node, g, var_def, ctxread=ctxread)

            ast_binop = ast.BinOp(left=left_ast, right=right_ast, op=g.get(n, 'ast_op'), lineno=0, col_offset=0)
            return ast_binop

        elif na == ast.UnaryOp:
            operand = g.get_edges_for_node_filter(end_node=n, attr='e_type', val=EdgeType.OPERAND)[1][0][0]

            operand_ast = node_to_ast(operand, g, var_def, ctxread=ctxread)

            ast_unop = ast.UnaryOp(operand=operand_ast, op=g.get(n, 'ast_op'), lineno=0, col_offset=0)
            return ast_unop

        elif na == ast.Call:

            args = [ii[0] for ii in g.get_edges_for_node_filter(end_node=n, attr='e_type', val=EdgeType.ARGUMENT)[1]]
            args_ast = []
            for a in args:
                a_ast = node_to_ast(a, g, var_def, ctxread=ctxread)
                args_ast.append(a_ast)

            ast_Call = ast.Call(args=args_ast, func=g.get(n, 'func'), keywords=[], lineno=0, col_offset=0)

            return ast_Call

        elif na == ast.IfExp:

            body = g.get_edges_for_node_filter(end_node=n, attr='e_type', val=EdgeType.BODY)[1][0][0]
            body_ast = node_to_ast(body, g, var_def, ctxread=ctxread)

            orelse = g.get_edges_for_node_filter(end_node=n, attr='e_type', val=EdgeType.ORELSE)[1][0][0]
            orelse_ast = node_to_ast(orelse, g, var_def, ctxread=ctxread)

            test = g.get_edges_for_node_filter(end_node=n, attr='e_type', val=EdgeType.TEST)[1][0][0]
            test_ast = node_to_ast(test, g, var_def, ctxread=ctxread)

            ast_ifexp = ast.IfExp(body=body_ast, orelse=orelse_ast, test=test_ast, lineno=0, col_offset=0)

            return ast_ifexp

        elif na == ast.Compare:
            comp = [ii[0] for ii in g.get_edges_for_node_filter(end_node=n, attr='e_type', val=EdgeType.COMP)[1]]
            comp_ast = []
            for a in comp:
                a_ast = node_to_ast(a, g, var_def, ctxread=ctxread)
                comp_ast.append(a_ast)

            left = g.get_edges_for_node_filter(end_node=n, attr='e_type', val=EdgeType.LEFT)[1][0][0]

            left_ast = node_to_ast(left, g, var_def, ctxread=ctxread)

            ast_Comp = ast.Compare(left=left_ast, comparators=comp_ast, ops=g.get(n, 'ops'), lineno=0, col_offset=0)

            return ast_Comp
        raise TypeError(f'Cannot convert {n},{na}')
    except:
        print(n)
        raise


def process_if_node(func_body, func_test):
    return ast.If(body=func_body, test=func_test, orelse=[], lineno=0, col_offset=0)


def process_assign_node(target_nodes, g, var_def, value_ast, na, targets):
    if len(target_nodes) > 1:
        target_ast = []
        for target_node in target_nodes:
            target_ast.append(node_to_ast(target_node[1], g, var_def, read=False))
            targets.append(target_node[1])
        ast_assign = ast.Assign(targets=[ast.Tuple(elts=target_ast, lineno=0, col_offset=0, ctx=ast.Store())],
                                value=value_ast, lineno=0, col_offset=0)
        return ast_assign
    else:
        target_node = target_nodes[0][1]
        target_ast = node_to_ast(target_node, g, var_def, read=False)
        if value_ast and target_ast:
            if na == ast.Assign or target_node not in targets:
                targets.append(target_node)
                ast_assign = ast.Assign(targets=[target_ast], value=value_ast, lineno=0, col_offset=0)
            else:
                ast_assign = ast.AugAssign(target=target_ast, value=value_ast, lineno=0, col_offset=0, op=ast.Add())
            return ast_assign


def generate_return_statement(var_def_, g):
    if (l := len(var_def_.get_targets())) > 1:
        return_ = ast.Return(value=ast.Tuple(elts=var_def_.get_order_trgs()))
    elif l == 1:
        return_ = ast.Return(value=var_def_.get_order_trgs()[0])
    else:
        # g.as_graphviz('noret', force=True)
        raise IndexError(f'Function {g.label} should have return, no?')
    return return_


def function_from_graph_generic(g: MappingsGraph, var_def_, arg_metadata, replacements=None, replace_name=None):
    decorators = []
    if replace_name is None:
        fname = g.label
    else:
        fname = replace_name
    body = function_body_from_graph(g, var_def_)
    var_def_.order_variables(arg_metadata)
    body.append(generate_return_statement(var_def_, g))

    args = ast.arguments(posonlyargs=[], args=var_def_.get_order_args(), vararg=None, defaults=[], kwonlyargs=[],
                         kwarg=None)

    func = wrap_function(fname, body, decorators=decorators, args=args)
    tree = replace_closure_function(func, replacements, get_eq_prefix())
    func_result = ast.parse(tree, mode='exec').body[0]
    target_ids = []
    for i, arg in enumerate(var_def_.args_order):
        if arg in var_def_.targets:
            target_ids.append(i)
    return func_result, var_def_.args_order, target_ids


def compare_expresion_from_graph(g, var_def_, lineno_count=1):
    top_nodes = g.topological_nodes()
    var_def = var_def_.var_def
    body = []
    for n in top_nodes:
        lineno_count += 1
        if g.get(n, 'ast_type') == ast.Compare:
            body.append(node_to_ast(n, g, var_def, ctxread=True))
    return body


def function_body_from_graph(g, var_def_, lineno_count=1, level=0):
    g.topological_nodes(ignore_cyclic=True)  # give warning if cyclic, but ignore sorting
    top_nodes = range(0, len(g.nodes))
    var_def = var_def_.var_def
    body = []
    targets = []
    for n in top_nodes:
        lineno_count += 1
        if (at := g.get(n, 'ast_type')) == ast.Assign or at == ast.AugAssign:
            value_node = g.get_edges_for_node_filter(end_node=n, attr='e_type', val=EdgeType.VALUE)[1][0][0]

            value_ast = node_to_ast(value_node, g, var_def, ctxread=True)
            body.append(process_assign_node(g.get_edges_for_node_filter(start_node=n, attr='e_type',
                                                                        val=EdgeType.TARGET)[1],
                                            g, var_def, value_ast, at, targets))
        if (g.get(n, 'ast_type')) == ast.If:
            func_body = function_body_from_graph(g.nodes[n].subgraph_body, var_def_,
                                                 lineno_count=lineno_count, level=level + 1)
            func_test = compare_expresion_from_graph(g.nodes[n].subgraph_test, var_def_,
                                                     lineno_count=lineno_count)
            body.append(process_if_node(func_body, func_test))
    return body


def replace_closure_function(func, replacements, eq_prefix):
    f1 = ast.unparse(func)
    if replacements:
        keys = list(replacements.keys())
        for key in keys:
            eq_key = (eq_prefix + key).replace('.', '_')
            f1 = f1.replace(key, eq_key)
            replacements[eq_key] = replacements[key]
            replacements.pop(key)
    return f1


def compiled_function_from_graph_generic_llvm(g: Graph, var_def_, imports,
                                              compiled_function=False, replacements=None, replace_name=None):
    func, signature, fname, r_args, r_targets = function_from_graph_generic_llvm(g, var_def_, replace_name=replace_name)
    if replacements is None:
        replacements = {}
    if not compiled_function:
        return func, signature, r_args, r_targets

    body = []
    for (module, g.label) in imports.as_imports:
        body.append(ast.Import(names=[ast.alias(name=module, asname=g.label)], lineno=0, col_offset=0, level=0))
    for (module, g.label) in imports.from_imports:
        body.append(
            ast.ImportFrom(module=module, names=[ast.alias(name=g.label, asname=None)], lineno=0, col_offset=0,
                           level=0))
    body.append(func)
    body.append(ast.Return(value=ast.Name(id=fname, ctx=ast.Load(), lineno=0, col_offset=0), lineno=0, col_offset=0))
    wrapper_name = fname + '1'

    func = wrap_function(wrapper_name, body, decorators=[],
                         args=ast.arguments(posonlyargs=[], args=[], vararg=None, defaults=[],
                                            kwonlyargs=[], kw_defaults=[], lineno=0, kwarg=None))

    tree = replace_closure_function(func, replacements, get_eq_prefix())
    tree = ast.parse(tree, mode='exec')
    module_func = ast.Module(body=[tree], type_ignores=[])
    code = compile(ast.parse(ast.unparse(module_func)), filename='llvm_equations_storage', mode='exec')
    namespace = replacements
    exec(code, namespace)
    compiled_func = namespace[wrapper_name]()

    return compiled_func, signature, r_args, r_targets


def function_from_graph_generic_llvm(g: Graph, var_def_, replace_name=None):
    if replace_name is None:
        fname = g.label + '_llvm'
    else:
        fname = replace_name + '_llvm'

    body = function_body_from_graph(g, var_def_)
    var_def_.order_variables(g.arg_metadata)
    args = ast.arguments(posonlyargs=[], args=var_def_.get_order_args(), vararg=None, defaults=[],
                         kwonlyargs=[], kw_defaults=[], kwarg=None)
    signature = [f'void(']
    target_ids = []
    for i, arg in enumerate(var_def_.args_order):
        if arg in var_def_.targets:
            signature.append("CPointer(float64), ")
            target_ids.append(i)
        else:
            signature.append("float64, ")
    signature[-1] = signature[-1][:-2]
    signature.append(")")
    signature = ''.join(signature)
    decorators = []

    func = wrap_function(fname, body, decorators=decorators, args=args)
    return func, signature, fname, var_def_.args_order, target_ids



from numerous.engine.system.binding import Binding
from numerous.utils.dict_wrapper import _DictWrapper
from numerous.engine.system.namespace import _ShadowVariableNamespace
from numerous.engine.system.node import Node



class Connector(Node):
    """
     Base class for representing connectors. Object that inherited connector can be used as
     a connection between items.

     Attributes
     ----------
          bindings :  dictionary of :class:`Binding`
               List of binding that connector have.
    """

    def __init__(self, tag, **kw):
        self.bindings = _DictWrapper(self.__dict__, Binding)
        super(Connector, self).__init__(tag)

    def create_binding(self, binding_name):
        """
        Creating a new binding inside the connector

        Parameters
        ----------
            binding_name : string
                name of the new binding

        Raises
        ------
        ValueError
            If `binding_name` is already registered in this connector.
        """
        if binding_name in self.bindings.keys():
            raise ValueError('Binding with name {0} is already registered in connector {1}'
                             .format(binding_name, self.tag))
        else:
            self.bindings[binding_name] = Connector.create_new_binding(binding_name)

    def _create_shadow_namespace(self, param):
        self.__update_bindings_namespace(param)

    def update_bindings(self, list_of_eq, binding_name):
        """
        Updating an existing binding with the equations that are expected to be in the binded items.

        Parameters
        ----------
        list_of_eq : list of :class:`numerous.multiphysics.Equation`
            List of a `Equation` that are expected to be in binded items.

        binding_name: string
            Name of a binding to be updated.
        """
        for binding in self.bindings:
            binding.__dict__[binding_name].add_equations(list_of_eq)

    def __update_bindings_namespace(self, param):
        for binding in self.bindings:
            binding.update_namespace(_ShadowVariableNamespace(self, param, binding))

    def get_binded_items(self):
        """
              Get items that are binded to the connector.

              Returns
              -------
              items : list
                  all items that are binded to the connector as one list.

        """
        return [
            y.binded_item for y in self.bindings
            if y.binded_item is not None
        ]

    @staticmethod
    def create_new_binding(binding_name):
        """
        Creates a new :class:`Binding` without registering it inside the connector.

        Parameters
        ----------
            binding_name: string
                Name of a binding to be created.

        Returns
        -------
            binding : Binding
                  new binding with given name.

        """
        return Binding(binding_name)

    def __setattr__(self, key, value):
        if key in self.__dict__:
            if isinstance(self.__dict__[key], Binding):
                if isinstance(value, Node):
                    self.__dict__[key].add_binding(value)
        object.__setattr__(self, key, value)



from enum import Enum


class SolverType(Enum):
    SOLVER_IVP = 0
    NUMEROUS = 1


solver_types = [SolverType.NUMEROUS, SolverType.SOLVER_IVP]


class BaseSolver:
    def __init__(self):
        pass



# #We can have multiple equation_function in same equation class. Since equation_functions that is decorated as
# differential equations should be treated differently we have two types of decorators.
import uuid
from functools import wraps
import inspect
from textwrap import dedent
from numba import njit

class NumerousFunction(object):
    def __init__(self, signature=None):
        self.id =str(uuid.uuid4())
        self.signature = signature

    def __call__(self, func):
        njited_func = njit(func)
        return njited_func






class Equation(object):

    def __init__(self):
        self.id =str(uuid.uuid4())

    def __call__(self, func):
        @wraps(func)
        def wrapper(self, scope):
            func(self, scope)

        if hasattr(func, '__self__'):
            wrapper.__self__ = func.__self__

        wrapper._equation = True
        wrapper.lines = inspect.getsource(func)
        wrapper.id = self.id
        #a = inspect.getsourcelines(func)
        wrapper.lineno = inspect.getsourcelines(func)[1]
        wrapper.file = inspect.getfile(func)
        wrapper.name = func.__name__
        self.name = func.__name__
        return wrapper


def add_equation(host, func):

    eq = Equation()
    eq_func = eq(func)

    host.equations.append(eq_func)


def dedent_code(code):
    tries = 0
    while tries < 10:
        try:
            dsource = dedent(dsource)

            break
        except IndentationError:

            tries += 1
            if tries > 10 - 1:
                print(dsource)
                raise

class InlineEquation(Equation):

    def __call__(self, func_name, func_source, namespace = {}):
        self.name = func_name
        tries = 0
        while tries < 10:
            try:
                func_source = dedent(func_source)
                exec(func_source, namespace)
                break
            except IndentationError:

                tries += 1
                if tries > 10 - 1:
                    print(func_source)
                    raise

        func = namespace[func_name]
        @wraps(func)
        def wrapper(self,scope):
            func(self, scope)
        wrapper.name = func_name
        wrapper._equation = True
        wrapper.lines = func_source
        wrapper.id = self.id
        #a = inspect.getsourcelines(func)
        wrapper.lineno = 0
        wrapper.file = 'dynamic.py'
        # wrapper.i = self.i
        return wrapper


class DifferentialEquation(Equation):

    def __init__(self, func):
        super().__init__(func)

    def __call__(self, scope):
        scope.apply_differential_equation_rules(True)
        self.func(scope)
        scope.apply_differential_equation_rules(False)



from numerous.engine.numerous_event import NumerousEvent, TimestampEvent
from numerous.engine.system.connector import Connector
from numerous.utils.dict_wrapper import _DictWrapper
from numerous.engine.system.namespace import VariableNamespace, VariableNamespaceBase
from numerous.engine.system.node import Node
from numerous.utils.logger_levels import LoggerLevel

DEFAULT_NAMESPACE = 'default'


class Item(Node):
    """
           Base class for items hierarchy. Contains fields and methods required to process
           item objects. An Item in a Numerous engine are represent any
           object needed for simulation model.


          Attributes
          ----------
          registered_namespaces :  dictionary of `VariableNamespaceBase`
              Dictionary of namespaces registered in the current item. All registered namespaces are although added to
              to the __dict__ of an object and can be referenced as a class attribute.

       """

    def __init__(self, tag=None, logger_level=None):
        self.registered_namespaces = _DictWrapper(self.__dict__, VariableNamespaceBase)
        self.events = []
        self.timestamp_events = []
        self.level = 1
        self.parent_item = None
        self.registered = False
        self.parent_set = None
        self.part_of_set = False

        self.logger_level = logger_level
        if tag:
            self.tag = tag

        self.path = [self.tag]
        super(Item, self).__init__(self.tag)

    def get_default_namespace(self):
        """
        Returns a new namespace with name default for the current item. This namespace is not
        registered.

        Returns
        -------
        namespace : `VariableNamespaceBase`
            a default namespace.

        Examples
        --------
        >>> item = Item('example')
        >>> dn = item.get_default_namespace()
        >>> print(dn.item is None )
        True
        >>> item.register_namespace(dn)
        >>> print(dn.item is None )
        False
        """
        return VariableNamespace(self, DEFAULT_NAMESPACE, is_connector=isinstance(self, Connector))

    def create_namespace(self, tag):
        """
        Creating a namespace.

        Parameters
        ----------
        tag : string
            Name of a class:`numerous.engine.VariableNamespace`

        Returns
        -------
        new_namespace : class:`numerous.engine.VariableNamespace`
            Empty namespace with given name

        """
        new_namespace = VariableNamespace(self, tag, is_connector=isinstance(self, Connector))
        self.register_namespace(new_namespace)
        return new_namespace

    def register_namespace(self, namespace):
        """
        Registering an already existed namespace for item.

        Parameters
        ----------
            namespace : `VariableNamespace`
                namespace to be registered.

        Raises
        ------
        ValueError
            If namespace is already registered for this item.
        """
        if not namespace.registered:
            if namespace.tag in self.registered_namespaces.keys():
                raise ValueError('Namespace with name {0} is already registered in item {1}'
                                 .format(namespace.tag, self.tag))
            else:
                self.registered_namespaces.update({namespace.tag: namespace})
            namespace.registered=True
        else:
            raise ValueError(f'Cannot register namespace {namespace.tag} more than once!')

    def get_variables(self):
        """
        Get variables from registered namespaces.

        Returns
        -------
        variables : list of tuples
            all variables with corresponding registered namespace. In for of tuple (variable,namespace).

        """
        variables_result = []
        for vn in self.registered_namespaces.values():
            for variable in vn.variables:
                variables_result.append((variable, vn))
        return variables_result

    def add_event(self, key, condition, action, terminal=True, direction=-1,compiled=False):
        condition = condition
        action = action
        event = NumerousEvent(key, condition, action, compiled, terminal, direction)
        self.events.append(event)

    def add_timestamp_event(self, key, action, timestamps):
        action = action
        event = TimestampEvent(key, action, timestamps)
        self.timestamp_events.append(event)

    def _increase_level(self):
        self.level = self.level+1

    def get_item(self, item_path):
        """
        Get an item using item path.

        Parameters
        ----------
        item_path: 'ItemPath'
            hierarchical path to an item inside the subsystem

        Returns
        -------
        Item : 'Item'
                returns an item found at given path or None

        """
        current_item = item_path.get_top_item()
        next_item_path = item_path.get_next_item_path()
        if self.tag == current_item and next_item_path:
            return None
        elif self.tag == current_item and next_item_path is None:
            return self
        else:
            return None

    def set_logger_level(self, logger_level=None):
        self.logger_level = logger_level



import copy
from enum import Enum

import networkx as nx

from numerous.multiphysics import EquationBase
from numerous.utils.dict_wrapper import _DictWrapper
from numerous.engine.system.item import Item
from numerous.engine.system.connector_item import ConnectorItem
from numerous.engine.system.namespace import SetNamespace


class ItemsStructure(Enum):
    LIST = 0
    GRID = 1
    SET = 2


class Subsystem(ConnectorItem):
    """
    Hierarchical representation of a group of items. Subsystems contains  a set of registered items, some
    of such items can be subsystems itself. This allow us to create Subsystems of any complexity.

    """

    def __init__(self, tag):
        self.ports = _DictWrapper({}, Item)
        self.registered_items = {}
        super().__init__(tag)

    def add_port(self, port_tag, item):
        """
        Creates a port for a subsystem. Ports can be used as an elements for binding.

        Parameters
        ----------
        port_tag: string
            Name of the port.
        item: :class:`numerous.engine.system.Item`
            item that is be used as a port.
        """

        if item.id not in self.registered_items.keys():
            raise ValueError(
                "Item {0} cannot be registered as port for subsystem {1}. Item is not registered in subsystem"
                    .format(item.tag, self.tag))
        self.ports[port_tag] = item

    def get_item(self, item_path):
        """
        Get an item using item path.

        Parameters
        ----------
        item_path: :class:`numerous.engine.system.ItemPath`
            hierarchical path to an item inside the subsystem

        Returns
        -------
        Item : :class:`numerous.engine.system.Item`
                returns an item found at given path or None

        """
        current_item = item_path.get_top_item()
        next_item_path = item_path.get_next_item_path()
        if self.tag == current_item and next_item_path:
            for item in self.registered_items.values():
                if item.tag == next_item_path.get_top_item():
                    return item.get_item(next_item_path)
            else:
                return None
        elif self.tag == current_item and next_item_path is None:
            return self
        else:
            return None

    def register_items(self, items, tag="set", structure=ItemsStructure.LIST, bind=False):
        """

        Parameters
        ----------
        items : list of :class:`numerous.engine.system.Item`
            List of items to register in the subsystem.
        """
        if structure == ItemsStructure.LIST:
            any(self.register_item(item) for item in items)
        elif structure == ItemsStructure.SET:
            itemset = ItemSet(items, tag)
            self.register_item(itemset)

        if bind:
            for i in items:
                setattr(self, i.tag, i)

    def increase_level(self):
        super().increase_level()
        for item in self.registered_items.values():
            item.increase_level()

    def get_graph_visualisation(self, DG=None, parent=None):
        if DG is None:
            DG = nx.DiGraph()
        DG.add_node(self.tag)
        if parent:
            DG.add_edge(parent, self.tag)
        for item in self.registered_items.values():
            if isinstance(item, Subsystem):
                item.get_graph_visualisation(DG, self.tag)
            else:
                DG.add_edge(self.tag, item.tag)
        return DG


    def update_variables_path(self, item, tail=[]):
        """
        Create variable path hierarchy.

        Parameters
        ----------
        item : :class:`numerous.engine.system.Item`
            Item to update variable path.
        tail:  list of :class:`numerous.engine.system.Item`
            In case of nested subsystems we store them to correctly update the path.

        """
        item.path = self.path + [item.tag]
        for ns in item.registered_namespaces.values():
            for var in ns.variables.values():
                var.path.extend_path(item.id, self.id, self.tag)
                if len(tail) == 1:
                    var.path.extend_path(self.id, tail[0].id, tail[0].tag)
                if len(tail) > 1:
                    t1 = tail[0::2][0]
                    t2 = tail[1::2][0]
                    var.path.extend_path(t2.id, t1.id, t1.tag)
        if isinstance(item, Subsystem):
            for item_ in item.registered_items.values():
                item.update_variables_path(item_, tail + [self])

    def register_item(self, item):
        """

        Parameters
        ----------
        item : :class:`numerous.engine.system.Item`
            Item to register in the subsystem.
        """
        if item.tag in [x.tag for x in self.registered_items.values()]:
            raise ValueError('Item with tag {} is already registered in system {}'.format(item.tag, self.tag))
        item._increase_level()
        self.update_variables_path(item)
        self.registered_items.update({item.id: item})


class ItemSet(Subsystem, EquationBase):

    def __init__(self, set_structure, tag):

        tag = "SET_" + tag
        super().__init__(tag)
        self.items = []
        set_structure_flat = set_structure

        self.item_ids = []

        self.item_type = None

        for item in set_structure_flat:
            item.part_of_set = True
            if not self.item_type:
                self.item_type = type(item)

            if not isinstance(item, self.item_type):
                raise TypeError(f'Error in registering set {tag} - '
                                f'All items in a set must have same type! This set is of type {self.item_type} '
                                f'not {type(item)}!')

            self.items.append(item)
            self.item_ids.append(item.id)
            if item.parent_set is None:
                item.parent_set = tag
            else:
                raise ValueError(f'Item {item} already part of set {item.parent_set} - cannot add to {tag}')

        tag_count = 0

        for item in set_structure_flat:
            for ns in item.registered_namespaces:
                tag_ = ns.tag
                if not (tag_ in self.registered_namespaces.keys()):
                    sns = SetNamespace(self, tag_, self.item_ids)
                    sns.add_equations(list(ns.associated_equations.values()), False, create_variables=False,
                                      set_equation=True)
                    sns.add_item(item, ns)
                    ns.clear_equations()
                    self.register_namespace(sns)
                else:
                    self.registered_namespaces[tag_].add_item(item, ns)
                if not ns.part_of_set:
                    ns.part_of_set = sns
                else:
                    ValueError(f'namespace {ns} already in set {ns.part_of_set}')
                tag_count += 1
        for ns in self.registered_namespaces.values():
            ns.items = self.items
        self.register_items(set_structure_flat)



import ast
from itertools import chain, zip_longest
from typing import Any

from numerous.engine.model.graph_representation.graph import Node, Edge, Graph
from numerous.engine.model.graph_representation.utils import str_to_edgetype, EdgeType
from numerous.engine.model.utils import NodeTypes, recurse_Attribute

op_sym_map = {ast.Add: '+', ast.Sub: '-', ast.Div: '/', ast.Mult: '*', ast.Pow: '**', ast.USub: '*-1',
              ast.Lt: '<', ast.LtE: '<=', ast.Gt: '>', ast.GtE: '>=', ast.Eq: '==', ast.NotEq: '!='}


def get_op_sym(op):
    return op_sym_map[type(op)]


def ast_to_graph(ast_tree, eq_key, eq_file, eq_lineno, scope_variables):
    parser = AstVisitor(eq_key, eq_file, eq_lineno, scope_variables)
    parser.visit(ast_tree)
    return parser.graph


class AstVisitor(ast.NodeVisitor):

    def __init__(self, eq_key, eq_file, eq_lineno, scope_variables):
        self.graph = Graph(label=eq_key)
        self.CONSTANT_LABEL = 'c'
        self.SCOPE_LABEL = 'scope.'
        self.IFEXPATTRIBUTES = ['body', 'orelse', 'test']
        self.branches = []
        self.eq_key = eq_key
        self.eq_file = eq_file
        self.eq_lineno = eq_lineno
        self.mapped_stack = []
        self.node_number_stack = []
        self.scope_variables = scope_variables
        self._supported_assign_target = (ast.Attribute, ast.Name, ast.Tuple)

    def traverse(self, node: ast.AST):
        if isinstance(node, list):
            for item in node:
                self.traverse(item)
        else:
            super().visit(node)

    def visit(self, node: ast.AST):
        self.traverse(node)

    def visit_Module(self, node: ast.Module):
        self.traverse(node.body)

    def visit_FunctionDef(self, node: ast.FunctionDef):
        self.traverse(node.body)

    def visit_Call(self, node: ast.Call) -> Any:
        self._process_named_node(node.func, ast_type=ast.Call, node_type=NodeTypes.OP, func=node.func)
        self.mapped_stack.pop()
        en = self.node_number_stack.pop()
        for sa in node.args:
            self.traverse(sa)
            start = self.node_number_stack.pop()
            self.mapped_stack.pop()
            self.graph.add_edge(Edge(start=start[0], end=en[0],
                                     e_type=EdgeType.ARGUMENT, branches=self.branches.copy()))
        self.mapped_stack.append([None])
        self.node_number_stack.append(en)

    def _is_assign_target_supported(self, target: ast.expr):
        return isinstance(target, self._supported_assign_target)

    def _process_named_node(self, node: ast.expr, ast_type: type, node_type: NodeTypes, static_key=False, func=None):
        source_id = recurse_Attribute(node)
        scope_var, is_mapped = self._select_scope_var(source_id)
        en = self.graph.add_node(
            Node(key=source_id if static_key else None, ao=node, file=self.eq_file, name=self.eq_key,
                 ln=self.eq_lineno, id=source_id, local_id=source_id, func=func,
                 ast_type=ast_type, node_type=node_type, scope_var=scope_var),
            ignore_existing=True)
        self.mapped_stack.append([is_mapped])
        self.node_number_stack.append([en])

    def visit_Assign(self, node: ast.Assign):
        for target in node.targets:
            if self._is_assign_target_supported(target):
                self.traverse(target)
            else:
                raise AttributeError('Unknown type of assign target in Equation: ', type(target))
        self.traverse(node.value)
        self.mapped_stack.pop()
        mapped = self.mapped_stack.pop()
        start = self.node_number_stack.pop()
        end = self.node_number_stack.pop()
        en = 0
        for s, e, m in zip_longest(start, end, mapped):
            if e is not None and s is not None:
                en = self.graph.add_node(Node(ao=node, file=self.eq_file, name=self.eq_key, ln=self.eq_lineno,
                                              label='+=' if m else '=',
                                              ast_type=ast.AugAssign if m else ast.Assign, node_type=NodeTypes.ASSIGN,
                                              ast_op=ast.Add() if m else None))
                self.graph.add_edge(Edge(start=en, end=e, e_type=EdgeType.TARGET, branches=self.branches.copy()))
                self.graph.add_edge(Edge(start=s, end=en, e_type=EdgeType.VALUE, branches=self.branches.copy()))
            else:
                if e is not None:
                    self.graph.add_edge(Edge(start=en, end=e, e_type=EdgeType.TARGET, branches=self.branches.copy()))
                if s is not None:
                    self.graph.add_edge(Edge(start=s, end=en, e_type=EdgeType.VALUE, branches=self.branches.copy()))
        self.node_number_stack.append([en])
        self.mapped_stack.append(mapped)

    def visit_Attribute(self, node: ast.Attribute):
        self._process_named_node(node, ast_type=ast.Attribute, static_key=True, node_type=NodeTypes.VAR)

    def visit_Name(self, node: ast.Name):
        self._process_named_node(node, ast_type=ast.Name, static_key=True, node_type=NodeTypes.VAR)

    #TODO: Enable Subscript visitation
    #def visit_Subscript(self, node: ast.Subscript):
    #    self._process_named_node(node, ast_type=ast.Subscript, static_key=True, node_type=NodeTypes.V)

    def _select_scope_var(self, source_id: str):
        scope_label_length = len(self.SCOPE_LABEL)
        if source_id[:scope_label_length] == self.SCOPE_LABEL:
            scope_var = self.scope_variables[source_id[scope_label_length:]]
            self.scope_variables[source_id[scope_label_length:]].used_in_equation_graph = True
            is_mapped = scope_var.sum_mapping or scope_var.mapping
        else:
            scope_var = None
            is_mapped = None
        return scope_var, is_mapped

    def visit_Constant(self, node: ast.Constant):
        source_id = self.CONSTANT_LABEL + str(node.value)
        en = self.graph.add_node(Node(ao=node, file=self.eq_file, name=self.eq_key, ln=self.eq_lineno,
                                      label=source_id, ast_type=ast.Num, value=node.value,
                                      node_type=NodeTypes.VAR))
        self.mapped_stack.append([None])
        self.node_number_stack.append([en])

    def visit_Tuple(self, node: ast.Tuple):
        en = []
        mapped = []
        for el in node.elts:
            self.traverse(el)
            en.append(self.node_number_stack.pop())
            mapped.append(self.mapped_stack.pop())
        self.mapped_stack.append(list(chain.from_iterable(mapped)))
        self.node_number_stack.append(list(chain.from_iterable(en)))

    def visit_BinOp(self, node: ast.BinOp):
        op_sym = get_op_sym(node.op)
        en = self.graph.add_node(Node(ao=node, file=self.eq_file, name=self.eq_key,
                                      ln=self.eq_lineno, label=op_sym, ast_type=ast.BinOp,
                                      node_type=NodeTypes.OP, ast_op=node.op))

        self.traverse(node.left)
        self.mapped_stack.pop()
        start = self.node_number_stack.pop()
        self.graph.add_edge(Edge(start=start[0], end=en,
                                 e_type=str_to_edgetype("left"), branches=self.branches.copy()))
        self.traverse(node.right)
        mapped = self.mapped_stack.pop()
        start = self.node_number_stack.pop()
        self.graph.add_edge(Edge(start=start[0], end=en,
                                 e_type=str_to_edgetype("right"), branches=self.branches.copy()))
        self.node_number_stack.append([en])
        self.mapped_stack.append(mapped)

    def visit_UnaryOp(self, node: ast.UnaryOp):

        op_sym = get_op_sym(node.op)
        en = self.graph.add_node(Node(label=op_sym, ast_type=ast.UnaryOp, node_type=NodeTypes.OP, ast_op=node.op),
                                 ignore_existing=True)
        self.traverse(node.operand)
        start = self.node_number_stack.pop()
        self.graph.add_edge(Edge(start=start[0], e_type=EdgeType.OPERAND, end=en, branches=self.branches.copy()))
        self.node_number_stack.append([en])

    def visit_Compare(self, node: ast.Compare) -> Any:
        ops_sym = [get_op_sym(o) for o in node.ops]
        en = self.graph.add_node(Node(ao=node, file=self.eq_file, name=self.eq_key, ln=self.eq_lineno,
                                      label=''.join(ops_sym), ast_type=ast.Compare,
                                      node_type=NodeTypes.OP, ops=node.ops))
        self.traverse(node.left)
        start = self.node_number_stack.pop()
        self.mapped_stack.pop()
        self.graph.add_edge(Edge(start=start[0], end=en, label=f'left', e_type=EdgeType.LEFT, branches=self.branches))
        for i, sa in enumerate(node.comparators):
            self.traverse(sa)
            start = self.node_number_stack.pop()
            mapped = self.mapped_stack.pop()
            self.graph.add_edge(
                Edge(start=start[0], end=en, label=f'comp{i}', e_type=EdgeType.COMP, branches=self.branches))
        self.mapped_stack.append(mapped)
        self.node_number_stack.append([en])

    def visit_IfExp(self, node: ast.IfExp) -> Any:
        en = self.graph.add_node(Node(ao=node, file=self.eq_file, name=self.eq_key, ln=self.eq_lineno, label='if_exp',
                                      ast_type=ast.IfExp, node_type=NodeTypes.OP))
        for a in self.IFEXPATTRIBUTES:
            self.traverse(getattr(node, a))
            start = self.node_number_stack.pop()
            mapped = self.mapped_stack.pop()
            self.graph.add_edge(Edge(start=start[0], end=en, e_type=str_to_edgetype(a), branches=self.branches))
        self.mapped_stack.append(mapped)
        self.node_number_stack.append([en])

    def visit_If(self, node: ast.If) -> Any:
        subgraph_body = ast_to_graph(node.body, self.eq_key, self.eq_file, self.eq_lineno, self.scope_variables)
        subgraph_test = ast_to_graph(node.test, self.eq_key, self.eq_file, self.eq_lineno, self.scope_variables)
        node = Node(ao=node, file=self.eq_file, name=self.eq_key,
                    ln=self.eq_lineno, label="IF", ast_type=ast.If, node_type=NodeTypes.IF, subgraph_body=subgraph_body,
                    subgraph_test=subgraph_test)
        subgraph_node_idx = connect_equation_node(subgraph_body, self.graph, node, is_set=False, include_local=True)
        connect_equation_node(subgraph_test, self.graph, node, is_set=False, include_local=True)
        self.mapped_stack.append([None])
        self.node_number_stack.append(subgraph_node_idx)


def connect_equation_node(equation_graph, mappings_graph, node, is_set, include_local=False):
    eq_node_idx = mappings_graph.add_node(node)
    for n in range(equation_graph.node_counter):
        if equation_graph.get(n, attr='node_type') == NodeTypes.VAR:
            if equation_graph.get(n, attr='scope_var'):
                sv = equation_graph.get(n, 'scope_var')
                neq = mappings_graph.add_node(Node(key=sv.id, node_type=NodeTypes.VAR, scope_var=sv,
                                                   is_set_var=is_set, label=sv.get_path_dot()),
                                              ignore_existing=True)

                targeted = False
                read = False

                end_edges = equation_graph.get_edges_for_node(end_node=n)

                try:
                    next(end_edges)
                    mappings_graph.add_edge(
                        Edge(start=eq_node_idx, end=neq, e_type=EdgeType.TARGET, arg_local=sv.id))
                    targeted = True
                except StopIteration:
                    pass

                if not targeted and not read:
                    start_edges = equation_graph.get_edges_for_node(start_node=n)
                    try:
                        next(start_edges)
                        mappings_graph.add_edge(Edge(neq, eq_node_idx, e_type=EdgeType.ARGUMENT, arg_local=sv.id))
                    except StopIteration:
                        pass

            elif include_local and equation_graph.get(n, attr='key') and \
                    not equation_graph.get(n, attr='ast_type') == ast.Num:
                var_key = equation_graph.get(n, attr='key')
                neq = mappings_graph.add_node(Node(key=var_key, node_type=NodeTypes.VAR,
                                                   is_set_var=is_set, ast_type=equation_graph.get(n, attr='ast_type'),
                                                   label=var_key),
                                              ignore_existing=True)

                targeted = False
                read = False

                end_edges = equation_graph.get_edges_for_node(end_node=n)

                try:
                    next(end_edges)
                    mappings_graph.add_edge(
                        Edge(start=eq_node_idx, end=neq, e_type=EdgeType.TARGET, arg_local='local'))
                    targeted = True
                except StopIteration:
                    pass

                if not targeted and not read:
                    start_edges = equation_graph.get_edges_for_node(start_node=n)
                    try:
                        next(start_edges)
                        mappings_graph.add_edge(Edge(neq, eq_node_idx, e_type=EdgeType.ARGUMENT, arg_local=sv.id if (
                            sv := equation_graph.get(n, 'scope_var')) else 'local'))
                    except StopIteration:
                        pass

    return eq_node_idx



class _DictWrapper(object):
    def __init__(self, internal_dict, value_type):
        self.value_type = value_type
        self.internal_dict = internal_dict
        self.shadow_dict = {}

    def keys(self):
        return self.shadow_dict.keys()

    def __getitem__(self, y):
        """ x.__getitem__(y) <==> x[y] """
        return self.shadow_dict[y]

    def __getattribute__(self, item):
        if item == 'shadow_dict' or item == '__setstate__' or item == '__dict__':
            return object.__getattribute__(self, item)
        if item in self.shadow_dict:
            return self.shadow_dict[item]
        return object.__getattribute__(self, item)

    def pop(self, item):
        if item in self.shadow_dict:
            self.shadow_dict.pop(item)

    def __iter__(self):
        return iter(self.shadow_dict.values())

    def values(self):
        return self.shadow_dict.values()

    def update(self, E=None, **F):
        self.internal_dict.update(E, **F)
        self.shadow_dict.update(E, **F)

    def __setitem__(self, key, value):
        if isinstance(value, self.value_type):
            self.shadow_dict[key] = value
            self.internal_dict[key] = value
        else:
            raise ValueError(
                "Object of type {0} cannot be added to internal dictionary of type {1}".format(type(value),
                                                                                               self.value_type))



import ast
import logging
from textwrap import dedent

from numerous.engine.variables import Variable
from numerous.engine.model.ast_parser.ast_visitor import ast_to_graph, connect_equation_node
from numerous.engine.model.graph_representation import Graph, EdgeType, Node, Edge
from numerous.engine.model.utils import NodeTypes
from copy import deepcopy


class ParsedEquation:
    def __init__(self, eq, dsource, graph, t, replacements=None):
        self.eq = eq
        self.dsource = dsource
        self.graph = graph
        self.t = t
        self.replacements = replacements


def attr_ast(attr):
    attr_ = attr.split('.')
    if len(attr_) > 1:
        prev = None
        attr_str = attr_[-1]
        attr_ = attr_[:-1]
        for a in attr_:
            if not prev:
                prev = ast.Name(id=a)
            else:
                prev = ast.Attribute(attr=a, value=prev)

        attr_ast = ast.Attribute(attr=attr_str, value=prev)
    else:
        attr_ast = ast.Name(id=attr_[0])
    return attr_ast


# Add nodes and edges to a graph
tmp_count = [0]


def tmp(a):
    a += '_' + str(tmp_count[0])
    tmp_count[0] += 1
    return a


ass_count = [0]


def ass(a):
    a += '_' + str(ass_count[0])
    ass_count[0] += 1
    return a


def postfix_from_branches(branches: dict):
    postfix = []
    for b, bv in branches.items():
        postfix += [b, str(bv)]
    return "_".join(postfix)


def qualify(s, prefix):
    return prefix + '.' + s.replace('scope.', '')


def qualify_equation(prefix, g, tag_vars, eq_class, eq_current):
    def q(s):
        return qualify(s, prefix)

    g_qual = g.clone()
    # update keys
    g_qual.node_map = {q(k): v for k, v in g_qual.node_map.items()}
    g_qual.key_map = {k: q(v) for k, v in g_qual.key_map.items()}

    scope_vars_qual = [tag_vars[sv.tag] if isinstance(sv := g.get(n, 'scope_var'), Variable) else sv
                       for n in g.node_map.values()]

    for i in range(g_qual.node_counter):
        g_qual.nodes[i].scope_var = scope_vars_qual[i]

    replacements = {}
    refer_to_self = False
    import inspect, hashlib, json

    closure_vars = inspect.getclosurevars(eq_class)

    func_ = closure_vars.nonlocals['func']
    closure_vars_func = inspect.getclosurevars(func_)
    replacements.update(closure_vars_func.globals)

    for n in g.nodes:

        if (f := n.func) is not None and hasattr(f, 'value'):
            if f.value.id == 'self':
                refer_to_self = True
                obj = getattr(eq_class.__self__, f.attr)
                replacements[n.id] = obj

    replacements_id = {k: id(o) for k, o in replacements.items()}

    eq_key = 'var_' + eq_current + '_' + hashlib.sha256(json.dumps(replacements_id).encode('UTF-8')).hexdigest()

    return g_qual, refer_to_self, eq_key, replacements


def _generate_equation_key(equation_id: str, is_set: bool) -> str:
    if is_set:
        eq_key = "EQ_SET" + equation_id.replace(".", "_").replace("-", "_")
    else:
        eq_key = "EQ_" + equation_id.replace(".", "_").replace("-", "_")
    return eq_key


def parse_eq(model_namespace, item_id, mappings_graph: Graph, scope_variables,
             parsed_eq_branches, scoped_equations, parsed_eq, eq_used):
    for m in model_namespace.equation_dict.values():
        for eq in m:
            ns_path = model_namespace.full_tag
            is_set = model_namespace.is_set
            eq_key = _generate_equation_key(eq.id, is_set)
            is_parsed_eq = eq_key in parsed_eq
            ast_tree = None
            if not is_parsed_eq:
                dsource = eq.lines

                tries = 0
                while tries < 10:
                    try:
                        dsource = dedent(dsource)
                        ast_tree = ast.parse(dsource)
                        break
                    except IndentationError:

                        tries += 1
                        if tries > 10 - 1:
                            raise

                g = ast_to_graph(ast_tree, eq_key, eq.file, eq.lineno, scope_variables)
                # Create branched versions of graph
                branches_ = set()
                [branches_.update(b.branches.keys()) for b in g.edges_c[:g.edge_counter] if b.branches]
                all_branches = [{}]
                for b in branches_:

                    for a in all_branches:
                        a.update({b: True})

                    all_branches += deepcopy(all_branches)
                    for a in all_branches[int(len(all_branches) / 2):]:
                        a[b] = False

                if len(all_branches) > 1:
                    branch_graphs = []
                    for a in all_branches:

                        gb = g.clone()

                        for i, b in enumerate(gb.edges_attr['branches'][:g.edge_counter]):

                            for ak in a.keys():
                                if ak in b and b[ak] != a[ak]:
                                    gb.remove_edge(i)

                        gb = gb.clean()
                        branch_graphs.append((a, gb, eq_key + '_' + postfix_from_branches(a)))

                    for branch in branch_graphs:
                        parsed_eq_branches[branch[2]] = ParsedEquation(eq, dsource, branch[1], branch[0])

                else:
                    parsed_eq_branches[eq_key] = ParsedEquation(eq, dsource, g, {})
                parsed_eq[eq_key] = 'EQ_' + ns_path + '.' + eq.name

            g = parsed_eq_branches[eq_key].graph

            eq_path = ns_path + '.' + eq_key
            g_qualified, refer_to_self, eq_key_, replacements = qualify_equation(ns_path, g, scope_variables, eq,
                                                                                 eq_key)

            if len(replacements) > 0:
                if refer_to_self:
                    eq_key__ = eq_key_
                else:
                    eq_key__ = eq_key

                parsed_eq_branches[eq_key__] = ParsedEquation(
                    parsed_eq_branches[eq_key].eq, parsed_eq_branches[eq_key].dsource, parsed_eq_branches[eq_key].graph,
                    {},
                    replacements)

                eq_key = eq_key__

            eq_used.append(eq_key)

            # make equation graph
            eq_name = ('EQ_' + eq_path).replace('.', '_')

            scoped_equations[eq_name] = eq_key

            node = Node(key=eq_name,
                        node_type=NodeTypes.EQUATION,
                        name=eq_name, file=eq_name, ln=0, label=eq_name,
                        ast_type=ast.Call,
                        vectorized=is_set,
                        item_id=item_id,
                        func=ast.Name(id=eq_key.replace('.', '_')))

            connect_equation_node(g_qualified, mappings_graph, node, is_set)

            if not is_parsed_eq:
                for sv in scope_variables:
                    if scope_variables[sv].used_in_equation_graph:
                        g.arg_metadata.append((sv, scope_variables[sv].id, scope_variables[sv].used_in_equation_graph))
                        scope_variables[sv].used_in_equation_graph = False
                    else:
                        g.arg_metadata.append((sv, scope_variables[sv].id, scope_variables[sv].used_in_equation_graph))


def process_mappings(mappings, mappings_graph: Graph, scope_vars):

    for m in mappings:
        target_var = scope_vars[m[0]]
        target_set_var_ix = -1
        if target_var.set_var:
            target_set_var_ix = target_var.set_var_ix
            target_var = target_var.set_var
        target_var_id = target_var.id

        node_type = NodeTypes.VAR

        t = mappings_graph.add_node(Node(key=target_var_id, file='mapping', name=m, ln=0, id=target_var_id,
                                         label=target_var.get_path_dot(), ast_type=ast.Attribute, node_type=node_type,
                                         scope_var=target_var, set_var_ix=target_set_var_ix), ignore_existing=False, )

        for i in m[1]:

            ivar_var = scope_vars[i]
            ivar_set_var_ix = -1

            if ivar_var.set_var:
                ivar_set_var_ix = ivar_var.set_var_ix
                ivar_var = ivar_var.set_var

            ivar_id = ivar_var.id

            ivar_node_e = mappings_graph.add_node(Node(key=ivar_id, file='mapping', name=m, ln=0, id=ivar_id,
                                                       label=ivar_var.get_path_dot(),
                                                       ast_type=ast.Attribute, node_type=NodeTypes.VAR,
                                                       scope_var=ivar_var,
                                                       set_var_ix=ivar_set_var_ix), ignore_existing=False)

            ix_ = mappings_graph.has_edge_for_nodes(start_node=ivar_node_e, end_node=t)
            lix = len(ix_)
            if lix == 0:
                mappings_graph.add_edge(Edge(ivar_node_e, t, e_type=EdgeType.MAPPING,
                                             mappings=[(ivar_set_var_ix, target_set_var_ix)]))
            else:
                mappings_graph.edges_c[ix_[0]].mappings.append((ivar_set_var_ix, target_set_var_ix))

    logging.info('Clone eq graph')

    logging.info('Remove dependencies')

    logging.info('Cleaning')

    mappings_graph = mappings_graph.clean()

    logging.info('Cleaning finished')
    return mappings_graph



import ast
import inspect

import numpy.typing as npt
import numpy as np
from numba.core.registry import CPUDispatcher

from numerous.engine.model.utils import njit_and_compile_function
from numerous.engine.numerous_event import NumerousEvent


def generate_event_condition_ast(event_functions: list[NumerousEvent],
                                 from_imports: list[tuple[str, str]]) -> tuple[list[CPUDispatcher], npt.ArrayLike]:
    array_label = "result"
    directions_array = []
    body = [ast.Assign(targets=[ast.Name(id=array_label)], lineno=0,
                       value=ast.List(elts=[], ctx=ast.Load()))]

    for event in event_functions:
        directions_array.append(event.direction)
        body.append(event.condition)
        body.append(ast.Expr(value=ast.Call(
            func=ast.Attribute(value=ast.Name(id=array_label, ctx=ast.Load()), attr='append', ctx=ast.Load()),
            args=[ast.Call(func=ast.Name(id=event.condition.name, ctx=ast.Load()),
                           args=[ast.Name(id='t', ctx=ast.Load()), ast.Name(id='states', ctx=ast.Load())],
                           keywords=[])], keywords=[])))

    body.append(ast.Return(value=ast.Call(func=ast.Attribute(value=ast.Name(id='np',
                                                                            ctx=ast.Load()), attr='array',
                                                             ctx=ast.Load()),
                                          args=[ast.Name(id=array_label, ctx=ast.Load()),
                                                ast.Attribute(value=ast.Name(id='np', ctx=ast.Load()), attr='float64',
                                                              ctx=ast.Load())],
                                          keywords=[])))
    body_r = ast.FunctionDef(name='condition_list',
                             args=ast.arguments(posonlyargs=[], args=[ast.arg(arg='t'), ast.arg(arg='states')],
                                                kwonlyargs=[], kw_defaults=[], defaults=[]),
                             body=body, decorator_list=[], lineno=0)

    return [njit_and_compile_function(body_r, from_imports)], np.array(directions_array)


def _replace_path_strings(model, function, idx_type, path_to_root=[]):
    lines = inspect.getsource(function)
    path_to_root_str = ".".join(path_to_root) + "."
    path_to_root_str_len = len(path_to_root_str)
    for (var_path, var) in model.path_to_variable.items():
        if var_path.startswith(path_to_root_str):
            var_path = var_path[path_to_root_str_len:]
        if var_path in lines:
            lines = lines.replace('[\'' + var_path + '\']', str(model._get_var_idx(var, idx_type)))
    func = ast.parse(lines.strip()).body[0]
    return func


def generate_event_action_ast(event_functions: list[NumerousEvent],
                              from_imports: list[tuple[str, str]]) -> CPUDispatcher:
    body = []

    for idx, event in enumerate(event_functions):
        body.append(event.action)
        body.append(ast.If(test=ast.Compare(left=ast.Name(id='a_idx', ctx=ast.Load()), ops=[ast.Eq()],
                                            comparators=[ast.Constant(value=idx)]),
                           body=[ast.Expr(value=ast.Call(func=ast.Name(id=event.action.name, ctx=ast.Load()),
                                                         args=[ast.Name(id='t', ctx=ast.Load()),
                                                               ast.Name(id='states', ctx=ast.Load())],
                                                         keywords=[]))], orelse=[]))

    body.append(ast.Return(value=ast.Name(id='states', ctx=ast.Load())))

    body_r = ast.FunctionDef(name='action_list',
                             args=ast.arguments(posonlyargs=[], args=[ast.arg(arg='t'), ast.arg(arg='states'),
                                                                      ast.arg(arg='a_idx')],
                                                kwonlyargs=[], kw_defaults=[], defaults=[]),
                             body=body, decorator_list=[], lineno=0)

    return njit_and_compile_function(body_r, from_imports)



from numba import njit, int64
from numba.experimental import jitclass
import numpy as np


@njit
def multi_replace(arr, to_rep, new_val):
    for t in to_rep:
        arr[:] = np.where(arr == t, new_val, arr)


@njit('i8(i8[:],i8)', cache=False)
def index(array, item):
    for ix, val in np.ndenumerate(array):
        if val == item:
            return int64(ix[0])
    return int64(-1)


@njit('Tuple((i8, i8[:]))(i8,i8[:],i8[:,:])', cache=False)
def depth_first_search(node, path, children):
    children_of_node = children[node, :]
    for i in range(children_of_node[0]):

        c = children_of_node[i + 1]
        if len(path) > 0:
            if index(path, c) >= 0:
                return c, np.append(path, c)

        path_ = np.append(path, c)

        cyclic_dependency, cyclic_path = depth_first_search(c, path_, children)
        if cyclic_dependency >= 0:
            return cyclic_dependency, cyclic_path

    return int64(-1), np.zeros((0,), dtype=int64)


@njit('Tuple((i8, i8, i8))(i8[:,:],i8[:,:],i8,i8[:,:],i8,i8[:],i8,i8[:], i8, i8[:])', cache=False)
def walk_parents(parent_edges, self_edges, n, edges, ix, visited_edges, n_visited, node_types, deriv_dep_count,
                 deriv_dep):
    for i in range(parent_edges[n, 0]):

        e = self_edges[parent_edges[n, i + 1]]

        if index(visited_edges[:n_visited], e[2]) < 0:

            visited_edges[n_visited] = e[2]
            n_visited += 1

            edges[ix, :] = e
            ix += 1
            if node_types[e[0]] < 3:

                ix, n_visited, deriv_dep_count = walk_parents(parent_edges, self_edges, e[0], edges, ix, visited_edges,
                                                              n_visited, node_types, deriv_dep_count, deriv_dep)
            elif node_types[e[0]] == 3:
                deriv_dep[deriv_dep_count] = e[0]
                deriv_dep_count += 1

    return ix, n_visited, deriv_dep_count


@njit('Tuple((i8, i8))(i8[:,:],i8[:,:],i8,i8[:,:],i8,i8[:],i8,i8[:])', cache=False)
def walk_parents_to_var(parent_edges, self_edges, n, edges, ix, visited_edges, n_visited, node_types):
    for i in range(parent_edges[n, 0]):
        e = self_edges[parent_edges[n, i + 1]]
        if index(visited_edges[:n_visited], e[2]) < 0:

            # if e[1] == n:
            visited_edges[n_visited] = e[2]
            n_visited += 1

            edges[ix, :] = e
            ix += 1

            if node_types[e[0]] < 2:
                ix, n_visited = walk_parents_to_var(parent_edges, self_edges, e[0], edges, ix, visited_edges, n_visited,
                                                    node_types)

    return ix, n_visited


@njit('Tuple((i8, i8))(i8[:,:],i8[:,:],i8[:,:],i8,i8[:,:],i8,i8[:],i8,i8[:])', cache=False)
def walk_children(parent_edges, children_edges, self_edges, n, edges, ix, visited_edges, n_visited, node_types):
    for i in range(children_edges[n, 0]):

        e = self_edges[children_edges[n, i + 1]]

        if index(visited_edges[:n_visited], e[2]) < 0:

            visited_edges[n_visited] = e[2]
            n_visited += 1

            edges[ix, :] = e
            ix += 1
            ix, n_visited = walk_children(parent_edges, children_edges, self_edges, e[1], edges, ix, visited_edges,
                                          n_visited, node_types)

            # if e[1] == n:
            if node_types[e[0]] < 2:
                ix, n_visited = walk_parents_to_var(parent_edges, self_edges, e[0], edges, ix, visited_edges, n_visited,
                                                    node_types)

    return ix, n_visited


@njit('Tuple((i8, i8))(i8[:,:],i8,i8[:,:],i8,i8[:],i8,i8[:])', cache=False)
def walk_parents_to_var_(self_edges, n, edges, ix, visited_edges, n_visited, node_types):
    for e in self_edges:
        if index(visited_edges[:n_visited], e[2]) < 0:

            if e[1] == n:
                visited_edges[n_visited] = e[2]
                n_visited += 1

                edges[ix, :] = e
                ix += 1

                if node_types[e[0]] < 2:
                    ix, n_visited = walk_parents_to_var_(self_edges, e[0], edges, ix, visited_edges, n_visited,
                                                         node_types)

    return ix, n_visited


@njit('Tuple((i8, i8))(i8[:,:],i8,i8[:,:],i8,i8[:],i8,i8[:])', cache=False)
def walk_children_(self_edges, n, edges, ix, visited_edges, n_visited, node_types):
    for e in self_edges:

        if index(visited_edges[:n_visited], e[2]) < 0:

            if e[0] == n:

                visited_edges[n_visited] = e[2]
                n_visited += 1

                edges[ix, :] = e
                ix += 1
                ix, n_visited = walk_children_(self_edges, e[1], edges, ix, visited_edges, n_visited, node_types)

                if node_types[e[0]] < 2:
                    ix, n_visited = walk_parents_to_var_(self_edges, e[0], edges, ix, visited_edges, n_visited,
                                                         node_types)

    return ix, n_visited


spec = [
    ('n_children_max', int64),
    ('n_nodes', int64),
    ('n_edges', int64),
    ('nodes', int64[:]),
    ('node_types', int64[:]),
    ('edges', int64[:, :]),
    ('children', int64[:, :]),
    ('parent_edges', int64[:, :]),
    ('children_edges', int64[:, :]),
    ('ancestors', int64[:, :]),
    ('indegree_map', int64[:]),
    ('topological_sorted_nodes', int64[:]),
    ('cyclic_dependency', int64),
    ('cyclic_path', int64[:])

]


@jitclass(spec)
class _Graph:
    def __init__(self, n_nodes: int64, edges: int64[:], node_types):

        self.n_children_max = 100
        self.n_nodes = n_nodes
        self.nodes = np.arange(self.n_nodes)

        self.node_types = node_types

        self.n_edges = len(edges)

        self.edges = edges

        self.indegree_map = np.zeros(self.n_nodes, np.int64)

        self.children = np.zeros((self.n_nodes, self.n_children_max), np.int64)

        self.parent_edges = np.zeros((self.n_nodes, self.n_children_max), np.int64)

        self.children_edges = np.zeros((self.n_nodes, self.n_children_max), np.int64)

        self.cyclic_dependency = np.int64(-1)
        self.cyclic_path = np.zeros((0,), np.int64)

        self.in_degree()

        self.make_children_map()
        self.topological_sorted_nodes = np.zeros(self.n_nodes, np.int64)

    def in_degree(self):

        for j in range(self.n_edges):
            self.indegree_map[self.edges[j][1]] += 1
        return self.indegree_map

    def make_children_map(self):

        for i, e in enumerate(self.edges):
            row = self.children[e[0], :]
            row[0] += 1
            if row[0] > self.n_children_max - 1:
                raise ValueError('More children than allowed!')
            row[row[0]] = e[1]

            row_pe = self.parent_edges[e[1], :]
            row_pe[0] += 1
            if row_pe[0] > self.n_children_max - 1:
                raise ValueError('More parents than allowed!')
            row_pe[row_pe[0]] = i

            row_ce = self.children_edges[e[0], :]
            row_ce[0] += 1
            if row_ce[0] > self.n_children_max - 1:
                raise ValueError('More parents than allowed!')
            row_ce[row_ce[0]] = i

    def make_edges_map(self, edges, nodes):
        n_nodes = len(nodes)
        parent_edges = np.zeros((n_nodes, self.n_children_max), int64)
        children_edges = np.zeros((n_nodes, self.n_children_max), int64)

        for i, e in enumerate(edges):

            row_pe = parent_edges[e[1], :]
            row_pe[0] += 1
            if row_pe[0] > self.n_children_max - 1:
                raise ValueError('More parents than allowed!')
            row_pe[row_pe[0]] = e[2]

            row_ce = children_edges[e[0], :]
            row_ce[0] += 1
            if row_ce[0] > self.n_children_max - 1:
                raise ValueError('More parents than allowed!')
            row_ce[row_ce[0]] = e[2]

        return parent_edges, children_edges

    def get_zero_indegree(self):
        n_zero_indegree = 0
        zero_indegree = np.zeros(self.n_nodes, dtype=int64)

        for i, im in enumerate(self.indegree_map):
            if im == 0:
                zero_indegree[n_zero_indegree] = i
                n_zero_indegree += 1
        return zero_indegree, n_zero_indegree

    def topological_sort(self):
        # print('n_nodes: ', self.n_nodes)
        sorted_nodes = np.zeros(self.n_nodes, int64) * -1
        n_sorted = 0

        zero_indegree, n_zero_indegree = self.get_zero_indegree()

        while n_zero_indegree > 0:
            # print(n_zero_indegree)
            node = zero_indegree[n_zero_indegree - 1]
            # print('zin: ',node)
            sorted_nodes[n_sorted] = node
            n_sorted += 1
            children_of_node = self.children[node, :]
            n_zero_indegree -= 1
            for i in range(children_of_node[0]):
                child = children_of_node[i + 1]
                self.indegree_map[child] -= 1
                if self.indegree_map[child] == 0:
                    zero_indegree[n_zero_indegree] = child
                    n_zero_indegree += 1

        if n_sorted < self.n_nodes:

            cd, self.cyclic_path = self.detect_cyclic_graph()

            self.cyclic_dependency = cd
        else:
            self.cyclic_dependency = -1

        self.topological_sorted_nodes = sorted_nodes

    def get_ancestor_graph(self, n):
        edges = np.zeros_like(self.edges)
        edges_visited = np.zeros(len(self.edges), dtype=int64)
        n_visited = int64(0)
        ix = int64(0)
        dep_derivatives = np.zeros(len(self.edges), dtype=int64)
        ix, n_visited, deriv_dep = walk_parents(self.parent_edges, self.edges, n, edges, ix, edges_visited, n_visited,
                                                self.node_types, 0, dep_derivatives)
        edges = edges[:ix, :]
        edges[:ix, 3] = 1

        nodes = np.zeros(2 * len(edges), dtype=int64)

        for i, e in enumerate(edges):
            nodes[2 * i] = e[0]
            nodes[2 * i + 1] = e[1]

        nodes = nodes[:i * 2 + 1 + 1]

        nodes = np.unique(nodes)

        return nodes, edges, dep_derivatives[:deriv_dep]

    def get_dependants_graph(self, nodes_):

        edges = np.zeros_like(self.edges)

        ix = int64(0)

        edges_visited = np.zeros(len(self.edges), dtype=int64)
        n_visited = int64(0)
        for n in nodes_:
            ix, n_visited = walk_children(self.parent_edges, self.children_edges, self.edges, n, edges, ix,
                                          edges_visited, n_visited, self.node_types)

        edges = edges[:ix, :]
        edges[:ix, 3] = 2

        nodes = np.zeros(2 * len(edges), dtype=int64)

        for i, e in enumerate(edges):
            nodes[2 * i] = e[0]
            nodes[2 * i + 1] = e[1]

        nodes = nodes[:i * 2 + 1 + 1]

        nodes = np.unique(nodes)

        return nodes, edges

    def get_dependants_graph_subgraph(self, nodes, subedges):

        edges = np.zeros_like(subedges)

        ix = int64(0)

        edges_visited = np.zeros(len(subedges), dtype=int64)
        n_visited = int64(0)
        for n in nodes:
            ix, n_visited = walk_children_(subedges, n, edges, ix, edges_visited, n_visited, self.node_types)

        edges = edges[:ix, :]
        edges[:ix, 3] = 2

        nodes = np.zeros(2 * len(edges), dtype=int64)

        for i, e in enumerate(edges):
            nodes[2 * i] = e[0]
            nodes[2 * i + 1] = e[1]

        nodes = np.unique(nodes)

        return nodes, edges

    def get_anc_dep_graph(self, n):
        anc_nodes, anc_edges, deriv_dependencies = self.get_ancestor_graph(n)
        state_nodes = np.array([ancn for ancn in anc_nodes if self.node_types[ancn] >= 3], np.int64)
        if len(state_nodes) > 0:
            nodes, edges = self.get_dependants_graph_subgraph(state_nodes, anc_edges)

        else:
            nodes = anc_nodes
            edges = anc_edges

        return nodes, edges, anc_nodes, anc_edges, deriv_dependencies

    def detect_cyclic_graph(self):
        zero_indegree, n_zero_indegree = self.get_zero_indegree()

        for i in range(n_zero_indegree):
            zid = zero_indegree[i]
            cyclic_dependency, cyclic_path = depth_first_search(zid, np.zeros((0,), dtype=int64), self.children)
            if cyclic_dependency >= 0:
                print('cyclic dependency: ', cyclic_dependency)
                print('cyclic path: ', cyclic_path)
                return cyclic_dependency, cyclic_path

        return int64(-1), np.zeros((0,), dtype=int64)



import ast
from enum import IntEnum, unique


class TemporaryKeyGenerator:
    def __init__(self):
        self.tmp_ = 0

    def generate(self):
        self.tmp_ += 1
        return f'tmp{self.tmp_}'


class Vardef:
    def __init__(self):
        self.vars_inds_map = []

    def var_def(self, var, read=True):
        if not var in self.vars_inds_map:
            self.vars_inds_map.append(var)
        ix = self.vars_inds_map.index(var)

        return ast.Subscript(slice=ast.Index(value=ast.Num(n=ix)), value=ast.Name(id='l'))


@unique
class EdgeType(IntEnum):
    TARGET = 0
    ARGUMENT = 1
    MAPPING = 2
    UNDEFINED = 3
    OPERAND = 4
    TMP = 5
    VALUE = 6
    LEFT = 7
    RIGHT = 8
    BODY = 9
    ORELSE = 10
    TEST = 11
    COMP = 12
    DEP = 13


def str_to_edgetype(a):
    if a == "left":
        return EdgeType.LEFT
    if a == "right":
        return EdgeType.RIGHT
    if a == 'body':
        return EdgeType.BODY
    if a == 'orelse':
        return EdgeType.ORELSE
    if a == 'test':
        return EdgeType.TEST



from dataclasses import dataclass
from enum import Enum
from typing import Any
import uuid
from functools import reduce
from operator import add
from numerous.utils.logger_levels import LoggerLevel


class VariableBase:
    pass


class VariableType(Enum):
    CONSTANT = 0
    PARAMETER = 1
    STATE = 2
    DERIVATIVE = 3
    PARAMETER_SET = 4
    TMP_PARAMETER = 5
    TMP_PARAMETER_SET = 6
    CALCULATED_VARIABLE = 7


class OverloadAction(Enum):
    RaiseError = 0
    SUM = 1


@dataclass
class VariableDescription:
    tag: str
    type: VariableType = VariableType.PARAMETER
    initial_value: Any = None
    id: str = None
    logger_level: LoggerLevel = LoggerLevel.ALL
    alias: str = None
    update: bool = False


@dataclass
class DetailedVariableDescription(VariableDescription):
    namespace: Any = None
    item: Any = None
    metadata: dict = None
    mapping: None = None
    update_counter: int = None
    allow_update: True = None


class MappedValue(object):
    def __init__(self, id):
        self.id = str(id).replace("-","_")
        self.mapping = None
        self.sum_mapping = []
        self.special_mapping = False
        self.addself = False
        self.logger_level = None
        self.model = None
        self.llvm_idx = None

    def add_mapping(self, variable):
        if not self.special_mapping:
            if variable.id == self.id:
                raise RecursionError("Variable {0} cannot be mapped to itself", self.id)
            self.mapping = variable
        self.special_mapping = False

    def add_sum_mapping(self, variable):
        self.sum_mapping.append(variable)

    def __iadd__(self, other):
        if isinstance(other, Variable):
            if self.mapping:
                raise ValueError('It is not possible to add a summation to {0}. Variable already have mapping'
                                 ''.format(self.tag))
            else:
                self.add_sum_mapping(other)
                self.special_mapping = True
                return self
        else:
            object.__iadd__(self, other)

    def __get_value(self, ids):
        if self.id in ids:
            return self.value
        else:
            if self.mapping:
                return self.mapping.get_value()
            if self.sum_mapping:
                ids.append(self.id)
                return reduce(add, [x.__get_value(ids) for x in self.sum_mapping])

            else:
                return self.value

    def get_value(self):
        if self.mapping:
            return self.mapping.get_value()
        if self.sum_mapping:
            return reduce(add, [x.__get_value([self.id]) for x in self.sum_mapping])

        else:
            return self.value

    def write_variable(self, value, llvm_idx):
        raise NotImplementedError


class VariablePath:

    def __init__(self, tag, id):
        self.path = {id: tag}
        self.primary_path = tag
        self.used_id_pairs = []

    def __iter__(self):
        return iter(self.path.values())

    def extend_path(self, current_id, new_id, new_tag):
        if not (current_id + new_id in self.used_id_pairs):
            if new_id in self.path:
                self.path[new_id].extend([new_tag + '.' + x for x in self.path[current_id]])
                self.primary_path = new_tag + '.' + self.path[current_id][-1]
            else:
                self.path.update({new_id: [new_tag + '.' + x for x in self.path[current_id]]})
                self.primary_path = new_tag + '.' + self.path[current_id][-1]
            self.used_id_pairs.append(current_id + new_id)


class SetOfVariables:
    def __init__(self, tag, item_tag, ns_tag):
        self.tag = tag
        self.id = "SET" + str(uuid.uuid4()).replace("-", "_")
        self.variables = {}
        self.mapping = []
        self.sum_mapping = []
        self.item_tag = item_tag
        self.ns_tag = ns_tag
        self.size = 0

    def get_size(self):
        return self.size

    def get_path_dot(self):
        result = list(self.variables.values())[0].get_path_dot()

        return result[:result.find(self.item_tag)] + self.item_tag + "." + self.ns_tag + "." + self.tag

    def add_variable(self, variable):
        self.variables.update({variable.id: variable})
        variable.set_var = self
        if variable.sum_mapping:
            self.sum_mapping.append(variable.sum_mapping)
        if variable.mapping:
            self.mapping.append(variable.mapping)
        self.size += 1

    def get_var_by_idx(self, i):
        return next(var for var in self.variables.values() if var.set_var_ix == i)

    def __iter__(self):
        return iter(self.variables.values())


class Variable(MappedValue):

    def __init__(self, detailed_variable_description, base_variable=None):

        super().__init__(detailed_variable_description.id)
        self.detailed_description = detailed_variable_description
        self.namespace = detailed_variable_description.namespace
        self.tag = detailed_variable_description.tag
        self.type = detailed_variable_description.type
        self.path = VariablePath([detailed_variable_description.tag], self.id)
        self.paths = []
        self.alias = None
        self.set_var = None
        self.set_var_ix = None
        self.set_namespace = None
        self.size = 0
        self.temporary_variable = False
        if base_variable:

            self._value = base_variable.value
        else:
            self._value = detailed_variable_description.initial_value
        self.item = detailed_variable_description.item
        self.metadata = detailed_variable_description.metadata
        self.mapping = detailed_variable_description.mapping
        self.update_counter = detailed_variable_description.update_counter
        self.allow_update = detailed_variable_description.allow_update
        self.logger_level = detailed_variable_description.logger_level
        self.associated_scope = []
        self.idx_in_scope = []
        self.top_item = None
        self.used_in_equation_graph = False

    def get_path_dot(self):
        return self.path.path[self.top_item][0]

    def update_value(self, value):
        self._value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        try:
            float(value)
        except ValueError:
            print(value)
        self._value = value
        if self.llvm_idx is not None:
            self.write_variable(value, self.llvm_idx)

    def update_set_var(self, set_var, set_namespace):
        if not self.set_var:
            self.set_var = set_var
            self.set_namespace = set_namespace
            print('path: ', self.get_path_dot())
            print('set_var: ', set_var)
            print('ns: ', set_namespace.tag)
        else:
            if self.set_var != set_var:
                print(self.set_var, ' ', set_var)
                raise ValueError(f'Setvar for {self.id} already set!')

    @staticmethod
    def create(namespace, v_id, tag,
               v_type, value, item, metadata,
               mapping, update_counter, allow_update, logger_level, alias):
        return Variable(DetailedVariableDescription(tag=tag,
                                                    id=v_id,
                                                    type=v_type,
                                                    initial_value=value,
                                                    namespace=namespace,
                                                    item=item,
                                                    metadata=metadata,
                                                    mapping=mapping,
                                                    update_counter=update_counter,
                                                    allow_update=allow_update,
                                                    logger_level=logger_level,
                                                    alias=alias))


class _VariableFactory:

    ##TODO remove recreation of var description here. It is duplicated inside the item
    @staticmethod
    def _create_from_variable_desc(namespace, item, var_desc):
        return Variable.create(namespace=namespace,
                               v_id="{0}_{1}_{2}_{3}".format(item.tag, namespace.tag, var_desc.tag, uuid.uuid4()),
                               tag=var_desc.tag,
                               v_type=var_desc.type,
                               value=var_desc.initial_value,
                               item=item,
                               metadata={},
                               mapping=None,
                               update_counter=0,
                               allow_update=(var_desc.type != VariableType.CONSTANT),
                               logger_level=var_desc.logger_level,
                               alias=var_desc.alias,
                               )

    @staticmethod
    def _create_from_variable_desc_unbound(initial_value, variable_description):
        v1 = Variable.create(namespace=None,
                             v_id="{0}_{1}".format(variable_description.tag, uuid.uuid4()),
                             tag=variable_description.tag,
                             v_type=variable_description.type,
                             value=initial_value,
                             item=None,
                             metadata={},
                             mapping=None,
                             update_counter=0,
                             allow_update=(variable_description.type != VariableType.CONSTANT),
                             logger_level=variable_description.logger_level,
                             alias=variable_description.alias,
                             )

        return v1



from numba import int32, float64, boolean, int64, njit, types, typed
import numpy as np
import numpy.typing as npt

# key and value types
kv_ty = (types.unicode_type, float64)

numba_model_spec = [
    ('start_time', int32),
    ('number_of_timesteps', int32),
    ('deriv_idx', int64[:]),
    ('state_idx', int64[:]),
    ('init_vars', float64[:]),
    ('global_vars', float64[:]),
    ('historian_ix', int64),
    ('historian_data', float64[:, :]),
    ('path_variables', types.DictType(*kv_ty)),
    ('path_keys', types.ListType(types.unicode_type)),
    ('historian_max_size', int64),
    ('in_memory_history_correction', boolean),
    ('external_mappings_time', float64[:, :]),
    ('number_of_external_mappings', int64),
    ('external_mappings_numpy', float64[:, :, :]),
    ('external_df_idx', int64[:, :]),
    ('approximation_type', boolean[:]),
    ('is_external_data', boolean),
    ('max_external_t', int64),
    ('external_idx', int64[:]),
]


@njit
def step_approximation(t, time_array, data_array):
    idx = np.searchsorted(time_array, t, side='right') - 1
    return data_array[idx]


class CompiledModel:
    def __init__(self, init_vars, deriv_idx, state_idx,
                 global_vars, number_of_timesteps, start_time, historian_max_size,
                 external_mappings_time, number_of_external_mappings,
                 external_mappings_numpy, external_df_idx, interpolation_info,
                 is_external_data, t_max, external_idx):
        self.external_idx = external_idx
        self.external_mappings_time = external_mappings_time
        self.number_of_external_mappings = number_of_external_mappings
        self.external_mappings_numpy = external_mappings_numpy

        self.external_df_idx = external_df_idx
        self.approximation_type = interpolation_info
        self.is_external_data = is_external_data
        self.max_external_t = t_max
        self.global_vars = global_vars
        self.deriv_idx = deriv_idx
        self.state_idx = state_idx
        self.path_variables = typed.Dict.empty(*kv_ty)
        self.path_keys = typed.List.empty_list(types.unicode_type)
        self.number_of_timesteps = number_of_timesteps
        self.start_time = start_time
        self.historian_ix = 0
        self.init_vars = init_vars
        self.historian_max_size = historian_max_size
        self.historian_data = np.empty(
            (len(init_vars) + 1, self.historian_max_size), dtype=np.float64)
        self.historian_data.fill(np.nan)

    def vectorizedfulljacobian(self, t, y, dt):
        h = 1e-8
        y_perm = y + h * np.diag(np.ones(len(y)))

        f = self.func(t, y)
        f_h = np.zeros_like(y_perm)
        for i in range(y_perm.shape[0]):
            y_i = y_perm[i, :]
            f_h[i, :] = self.func(t, y_i)

        diff = f_h - f
        diff /= h
        jac = diff.T
        return np.ascontiguousarray(jac)

    def map_external_data(self, t):
        for i in range(self.number_of_external_mappings):
            df_indx = self.external_df_idx[i][0]
            var_idx = self.external_df_idx[i][1]
            value = np.interp(t, self.external_mappings_time[df_indx],
                              self.external_mappings_numpy[df_indx, :, var_idx]) if \
                self.approximation_type[i] else \
                step_approximation(t, self.external_mappings_time[df_indx],
                                   self.external_mappings_numpy[df_indx, :, var_idx])
            self.write_variables(value, self.external_idx[i])

    def update_external_data(self, external_mappings_numpy, external_mappings_time):
        self.external_mappings_time = external_mappings_time
        self.external_mappings_numpy = external_mappings_numpy

    def is_external_data_update_needed(self, t):
        if self.is_external_data and t > self.max_external_t:
            return True
        return False

    def get_states(self):
        return self.read_variables()[self.state_idx]

    def set_states(self, states: npt.ArrayLike) -> None:
        for i in range(len(states)):
            self.write_variables(states[i], self.state_idx[i])

    def is_store_required(self):
        if self.historian_ix >= self.historian_max_size:
            return True
        return False

    def historian_reinit(self):
        self.historian_data = np.empty(
            (len(self.init_vars) + 1, self.historian_max_size), dtype=np.float64)
        self.historian_data.fill(np.nan)
        self.historian_ix = 0

    def historian_update(self, time: np.float64) -> None:
        ix = self.historian_ix
        self.historian_data[0][ix] = time
        self.historian_data[1:, ix] = self.read_variables()
        self.historian_ix += 1

    def get_g(self, t, yold, y, dt, order, a, af):
        f = self.func(t, y)
        _sum = np.zeros_like(y)
        for i in range(order):
            _sum = _sum + a[order - 1][i] * yold[i, :]
        g = y + _sum - af[order - 1] * dt * f
        return np.ascontiguousarray(g), np.ascontiguousarray(f)

    def func(self, _t, y):
        self.global_vars[0] = _t
        deriv = self.compiled_compute(y)
        return deriv



from __future__ import print_function

import os
import types

from numerous.engine.model.utils import wrap_function
from numerous.utils import config
import numpy as np
import ast
from copy import deepcopy as copy

from numerous.engine.model.lowering.utils import generate_code_file

GLOBAL_ARRAY = ast.Name(id="kernel_variables")
LISTING_FILEPATH = "tmp/listings/"
LISTINGFILENAME = "_kernel.py"


class ASTBuilder:
    """
    Building an AST module.
    """

    def __init__(self, initial_values, variable_names, states, derivatives, system_tag=""):
        """
        initial_values - initial values of global variables array. Array should be ordered in  such way
         that all the derivatives are located in the tail.
        variable_names - dictionary
        states - states
        derivatives -  derivatives
        """
        self.body_init_set_var = [ast.Assign(targets=[GLOBAL_ARRAY],
                                             value=ast.Call(func=ast.Name(id='np.array',
                                                                          ctx=ast.Load()),
                                                            args=[ast.List(
                                                                elts=[ast.Constant(value=v) for v in initial_values],
                                                                ctx=ast.Load())],
                                                            keywords=[]
                                                            ), lineno=0)]

        self.kernel_function = []
        self.variable_names = variable_names
        self.states = states
        self.derivatives = derivatives
        self.functions = []
        self.local_functions = {}
        self.defined_functions = []
        self.body = []
        self.replacements = []
        self.replace_name = []
        self.kernel_filename = LISTING_FILEPATH + system_tag + LISTINGFILENAME
        self.cnd_calls={}

        self.read_args_section = [ast.Expr(value=ast.Call(func=ast.Name(id='np.put'),
                                                          args=[GLOBAL_ARRAY,
                                                                ast.List(
                                                                    elts=[ast.Constant(value=v) for v in
                                                                          [self.variable_names[x] for x in
                                                                           self.states]],
                                                                    ctx=ast.Load()),
                                                                ast.Name(id='states', ctx=ast.Store())
                                                                ],
                                                          keywords=[]))]

        self.return_section = [ast.Return(
            value=ast.Call(func=ast.Name(id='np.take'), args=[GLOBAL_ARRAY,
                                                              ast.List(
                                                                  elts=[ast.Constant(value=v) for v in
                                                                        [self.variable_names[x] for x in
                                                                         self.derivatives]],
                                                                  ctx=ast.Load())
                                                              ],
                           keywords=[]))]

    def add_external_function(self, function: ast.FunctionDef, signature: str, number_of_args: int,
                              target_ids: list[int], replacements=None, replace_name=None):
        if replacements is not None:
            self.replacements.append(replacements)
            self.replace_name.append(replace_name)
        self.functions.append(function)
        self.defined_functions.append(function.name)

    def generate(self, imports, system_tag="", external_functions_source=False, save_to_file=False):
        arguments = [ast.arg(arg="states", annotation=None)]
        defaults=[]

        for i in self.body:
            if hasattr(i, 'cnd') and i.cnd:
                arguments.append(ast.arg(arg=i.body[0].value.func.id+'_flag', annotation=None))
                defaults.append(ast.Constant(i.active))

        kernel = wrap_function('global_kernel', self.read_args_section + self.body + self.return_section,
                               decorators=[],
                               args=ast.arguments(posonlyargs=[], args=arguments, vararg=None,
                                                  defaults=defaults,
                                                  kwarg=None, kwonlyargs=[]))
        variable_names_print = []
        for key, value in self.variable_names.items():
            variable_names_print.append('#' + str(key) + ' : ' + str(value))
        code = generate_code_file([x for x in self.functions] + self.body_init_set_var + [kernel], self.kernel_filename,
                                  imports,
                                  external_functions_source=external_functions_source,
                                  names='\n'.join(variable_names_print) + '\n')

        kernel_module = types.ModuleType('python_kernel')
        if self.replacements:
            local_replacments = {k: v for d in self.replacements for k, v in d.items()}
            kernel_module.__dict__.update(local_replacments)
        if save_to_file:
            os.makedirs(os.path.dirname(self.kernel_filename), exist_ok=True)
            with open(self.kernel_filename, 'w') as f:
                f.write(code)
            exec('from tmp.listings.' + system_tag + '_kernel import *', globals())

            def var_func():
                return kernel_variables

            def var_write(value, idx):
                np.put(kernel_variables, [idx], value)

            return global_kernel, var_func, var_write
        else:
            exec(code, kernel_module.__dict__)

            def var_func():
                return kernel_module.kernel_variables

            def var_write(value, idx):
                np.put(kernel_module.kernel_variables, [idx], value)

            return kernel_module.global_kernel, var_func, var_write,

        return kernel_module.global_kernel, var_func, var_write

    def unparse(self, imports, system_tag="", external_functions_source=False):
        arguments = [ast.arg(arg="states", annotation=None)]
        defaults=[]

        for i in self.body:
            if hasattr(i, 'cnd') and i.cnd:
                arguments.append(ast.arg(arg=i.body[0].value.func.id+'_flag', annotation=None))
                defaults.append(ast.Constant(i.active))

        kernel = wrap_function('global_kernel', self.read_args_section + self.body + self.return_section,
                               decorators=[],
                               args=ast.arguments(posonlyargs=[], args=arguments, vararg=None,
                                                  defaults=defaults,
                                                  kwarg=None, kwonlyargs=[]))
        variable_names_print = []
        for key, value in self.variable_names.items():
            variable_names_print.append('#' + str(key) + ' : ' + str(value))
        code = generate_code_file([x for x in self.functions] + self.body_init_set_var + [kernel], self.kernel_filename,
                                  imports,
                                  external_functions_source=external_functions_source,
                                  names='\n'.join(variable_names_print) + '\n')

        return code

    def detailed_print(self, *args, sep=' ', end='\n', file=None):
        if config.PRINT_LLVM:
            print(*args, sep, end, file)

    def store_variable(self, variable_name):
        pass

    def _create_assignments(self, external_function_name, input_args, target_ids):
        arg_ids = map(lambda arg: self.variable_names[arg], input_args)
        targets = []
        args = []
        for target_id in target_ids:
            targets.append(
                ast.Subscript(value=GLOBAL_ARRAY,
                              slice=ast.Index(value=ast.Constant(value=self.variable_names[input_args[target_id]])),
                              ctx=ast.Store()))
        for arg_id in arg_ids:
            args.append(
                ast.Subscript(value=GLOBAL_ARRAY, slice=ast.Index(value=ast.Constant(value=arg_id)),
                              ctx=ast.Load))

        if len(targets) > 1:
            temp = (ast.Assign(targets=[ast.Tuple(elts=targets)],
                               value=ast.Call(func=ast.Name(id=external_function_name, ctx=ast.Load()),
                                              args=args, keywords=[]), lineno=0))
        else:
            temp = (ast.Assign(targets=[targets[0]],
                               value=ast.Call(func=ast.Name(id=external_function_name, ctx=ast.Load()),
                                              args=args, keywords=[]), lineno=0))
        return temp

    def add_conditional_call(self,external_function_name, input_args, target_ids, tag, active=True):
        active=active

        arg_ids = map(lambda arg: self.variable_names[arg], input_args)
        targets=[]
        args=[]

        for target_id in target_ids:
            targets.append(
                ast.Subscript(value=GLOBAL_ARRAY,
                              slice=ast.Index(value=ast.Constant(value=self.variable_names[input_args[target_id]])),
                              ctx=ast.Store()))

        for arg_id in arg_ids:
            args.append(
                ast.Subscript(value=GLOBAL_ARRAY, slice=ast.Index(value=ast.Constant(value=arg_id)),
                              ctx=ast.Load))

        if len(targets) > 1:
            targets=[ast.Tuple(elts=targets)]
        else:
            targets=[targets[0]]

        ast_condition = ast.Name(id=external_function_name + '_flag', ctx=ast.Load())

        temp=ast.If(test=ast_condition, body=[
            ast.Assign(targets=targets, value=
                ast.Call(func=ast.Name(id=external_function_name, ctx=ast.Load()),
                                               args=args, keywords=[]),
                lineno=0)],
            orelse=[])



        setattr(temp, 'cnd', True)
        setattr(temp, 'active', active)
        self.cnd_calls.update({tag: len(self.body)})
        self.body.append(temp)

    def get_call_enabled(self, tag):
        return self.body[self.cnd_calls[tag]].active

    def set_call_enabled(self, tag, enabled):
        self.body[self.cnd_calls[tag]].active=enabled

    def add_call(self, external_function_name, input_args, target_ids):
        temp = self._create_assignments(external_function_name,input_args,target_ids)
        setattr(temp, 'cnd', False)
        self.body.append(temp)

    def add_mapping(self, args, targets):
        for target in targets:
            target = [target]
            if len(target) > 1:
                raise ValueError("Only mapping to single target is supported")
            arg_idxs = []
            for arg in args:
                arg_idxs.append(self.variable_names[arg])
            target_idx = self.variable_names[target[0]]
            print(target_idx)
            if len(args) == 1:
                temp=(ast.Assign(targets=[ast.Subscript(value=GLOBAL_ARRAY,
                                                                   slice=ast.Index(
                                                                       value=ast.Constant(value=target_idx,
                                                                                          kind=None)))],
                                            value=ast.Subscript(value=GLOBAL_ARRAY,
                                                                slice=ast.Index(
                                                                    value=ast.Constant(value=arg_idxs[0], kind=None)))
                                            , lineno=0))
            else:
                temp=(ast.Assign(targets=[ast.Subscript(value=GLOBAL_ARRAY,
                                                                   slice=ast.Index(
                                                                       value=ast.Constant(value=target_idx,
                                                                                          kind=None)))],
                                            value=ast.BinOp(left=self._generate_sum_left(arg_idxs[1:]), op=ast.Add(),
                                                            right=ast.Subscript(value=GLOBAL_ARRAY,
                                                                                slice=ast.Index(
                                                                                    value=ast.Constant(
                                                                                        value=arg_idxs[0],
                                                                                        kind=None))))
                                            , lineno=0))
            setattr(temp, 'cnd', False)

            self.body.append(temp)

    def _generate_sum_left(self, arg_idxs):
        if len(arg_idxs) > 0:
            return ast.BinOp(left=self._generate_sum_left(arg_idxs[1:]), op=ast.Add(),
                             right=ast.Subscript(value=GLOBAL_ARRAY,
                                                 slice=ast.Index(
                                                     value=ast.Constant(value=arg_idxs[0], kind=None))))
        else:
            ## adding 0 if no summation arguments left
            return ast.Constant(value=0, kind=None)

    def add_set_call(self, external_function_name, variable_name_arg_and_trg, targets_ids):
        temp=(ast.For(target=ast.Name(id='i', ctx=ast.Store()),
                                 iter=ast.Call(func=ast.Name(id='range', ctx=ast.Load()), args=
                                 [ast.Constant(value=len(variable_name_arg_and_trg))], keywords=[]),
                                 body=[self._generate_set_call_body(external_function_name,
                                                                    len(variable_name_arg_and_trg[0]),
                                                                    self.variable_names[
                                                                        variable_name_arg_and_trg[0][0]],
                                                                    targets_ids)],
                                 orelse=[], lineno=0))
        setattr(temp, 'cnd', False)
        self.body.append(temp)
    def _generate_set_call_body(self, external_function_name, arg_length, strart_idx, target_ids):
        arg_ids = np.arange(arg_length)
        targets = []
        args = []

        for target_id in target_ids:
            targets.append(ast.Subscript(value=GLOBAL_ARRAY,
                                         slice=ast.Index(
                                             value=ast.BinOp(
                                                 left=ast.Constant(value=target_id + strart_idx),
                                                 op=ast.Add(),
                                                 right=ast.BinOp(left=ast.Constant(value=arg_length, kind=None),
                                                                 op=ast.Mult(),
                                                                 right=ast.Name(id='i', ctx=ast.Load())))),
                                         ctx=ast.Store()))

        for arg_id in arg_ids:
            args.append(
                ast.Subscript(value=GLOBAL_ARRAY, slice=ast.Index(value=ast.BinOp(
                    left=ast.Constant(value=arg_id + strart_idx),
                    op=ast.Add(),
                    right=ast.BinOp(left=ast.Constant(value=arg_length, kind=None),
                                    op=ast.Mult(),
                                    right=ast.Name(id='i', ctx=ast.Load())))),
                              ctx=ast.Load))
        if len(targets) > 1:
            return ast.Assign(targets=[ast.Tuple(elts=targets)],
                              value=ast.Call(func=ast.Name(id=external_function_name, ctx=ast.Load()),
                                             args=args, keywords=[]), lineno=0)
        else:
            return ast.Assign(targets=[targets[0]],
                              value=ast.Call(func=ast.Name(id=external_function_name, ctx=ast.Load()),
                                             args=args, keywords=[]), lineno=0)

    def add_set_mapping(self, args2d, targets2d):
        pass



from numerous.utils.dict_wrapper import _DictWrapper
from .namespace import VariableNamespaceBase


class _NamespaceManagerList:
    def __init__(self, ns):
        self.__dict__[ns.tag] = ns

    def add_namespace(self, ns):
        self.__dict__[ns.tag] = ns


class Binding:
    """
        Represents expected binding of an `Item`. When binding is point to an actual item it checks for
        mapped variables and namespaces.

        Attributes
        ----------
             binding_name :  string
    """

    def __init__(self, binding_name):
        self.name = binding_name
        self.ns = _DictWrapper(self.__dict__, VariableNamespaceBase)
        self.binded_item = None

    def update_namespace(self, ns):
        """
            Updating an existing or creating a new namespace.

            Parameters
            ----------
            ns : :class:`Item`
                Namespace.
        """
        self.ns[ns.tag] = ns

    def __create_binding_varaible_bindings(self, namespace):
        if namespace.tag in self.binded_item.registered_namespaces.keys():
            for bv in namespace.variables:
                if bv.mapping:
                    self.binded_item.registered_namespaces[namespace.tag].variables[bv.tag].add_mapping(bv.mapping)
                if bv.sum_mapping:
                    for v in bv.sum_mapping:
                        self.binded_item.registered_namespaces[namespace.tag].variables[bv.tag].add_sum_mapping(v)

    def __merge_namespaces(self):
        for namespace in self.ns:
            self.__create_binding_varaible_bindings(namespace)
            self.ns[namespace.tag] = self.binded_item.registered_namespaces[namespace.tag]

    def add_binding(self, item):
        """
            Add an 'Item' to binded items.

            Parameters
            ----------
            item : :class:`Item`
                Item to be used as a binding.
        """
        if self.binded_item:
            if self.binded_item.id == item.id:
                raise ValueError("item {0} is already binded to binding {1}".format(item.tag, self.name))
            else:
                raise ValueError("Binding multiple items to the same binding is not supported")
        else:
            self.binded_item = item
            self.__merge_namespaces()

    def is_bindend(self):
        """
            Checks if item is binded to this binding.

            Returns
            -------
            is_bindend : bool
              """
        return not (self.binded_item is None)


from __future__ import print_function

import logging
import os
from ctypes import CFUNCTYPE, POINTER, c_double, c_void_p, c_int64
from numba import carray, cfunc, njit
from numerous.utils import config
import faulthandler
import numpy as np

import llvmlite.ir as ll
import llvmlite.binding as llvm

faulthandler.enable()
llvm.initialize()
llvm.initialize_native_target()
llvm.initialize_native_asmprinter()
llvmmodule = llvm.parse_assembly("")
target_machine = llvm.Target.from_default_triple().create_target_machine()
ee = llvm.create_mcjit_compiler(llvmmodule, target_machine)

LISTING_FILEPATH = "tmp/listings/"
LISTINGFILENAME = "_llvm_listing.ll"


class LLVMBuilder:
    """
    Building an LLVM module.
    """

    def __init__(self, initial_values, variable_names, states, derivatives, system_tag=""):
        """
        initial_values - initial values of global variables array. Array should be ordered in  such way
         that all the derivatives are located in the tail.
        variable_names - dictionary
        states - states
        derivatives -  derivatives
        """

        self.loopcount = 0
        self.detailed_print('Target data: ', target_machine.target_data)
        self.detailed_print('Target triple: ', target_machine.triple)
        self.module = ll.Module()
        self.ext_funcs = {}
        self.states = states
        self.derivatives = derivatives
        self.n_deriv = len(derivatives)
        # Define the overall function
        self.fnty = ll.FunctionType(ll.DoubleType().as_pointer(), [
            ll.DoubleType().as_pointer()
        ])
        self.index0 = 0
        self.fnty.args[0].name = "y"
        self.system_tag = system_tag
        self.variable_names = {}
        self.equations_llvm = []
        self.equations_llvm_opt = []
        for k, v in variable_names.items():
            self.variable_names[k] = v
            self.variable_names[v] = k

        self.max_var = initial_values.shape[0]

        self.var_global = ll.GlobalVariable(self.module, ll.ArrayType(ll.DoubleType(), self.max_var), 'global_var')
        self.var_global.initializer = ll.Constant(ll.ArrayType(ll.DoubleType(), self.max_var),
                                                  [float(v) for v in initial_values])

        self.deriv_global = ll.GlobalVariable(self.module, ll.ArrayType(ll.DoubleType(), self.n_deriv), 'derivatives')
        self.deriv_global.initializer = ll.Constant(ll.ArrayType(ll.DoubleType(), self.n_deriv),
                                                    [float(0) for _ in range(self.n_deriv)])

        # Creatinf the fixed structure of the kernel
        self.func = ll.Function(self.module, self.fnty, name="kernel")
        self.bb_entry = self.func.append_basic_block(name='entry')
        self.bb_loop = self.func.append_basic_block(name='main')
        self.bb_store = self.func.append_basic_block(name='store')
        self.bb_exit = self.func.append_basic_block(name='exit')
        self.builder = ll.IRBuilder()
        self.builder.position_at_end(self.bb_entry)
        self.index0 = ll.IntType(64)(0)

        self.values = {}

        # go through  all states to put them in load block
        for idx, state in enumerate(self.states):
            self.load_state_variable(idx, state)

        # go through all states to put them in store block
        for state in self.states:
            self.store_variable(state)

        self.builder.position_at_end(self.bb_entry)
        self.builder.branch(self.bb_loop)
        self.builder.position_at_end(self.bb_loop)

    def add_external_function(self, function, signature, number_of_args, target_ids, replacements=None,
                              replace_name=None):
        """
        Wrap the function and make it available in the LLVM module
        """
        f_c = cfunc(sig=signature)(function)
        name = f_c.native_name
        self.equations_llvm.append(f_c.inspect_llvm())

        f_c_sym = llvm.add_symbol(name, f_c.address)
        llvm_signature = np.tile(ll.DoubleType(), number_of_args).tolist()
        for i in target_ids:
            llvm_signature[i] = ll.DoubleType().as_pointer()

        fnty_c_func = ll.FunctionType(ll.VoidType(),
                                      llvm_signature)
        fnty_c_func.as_pointer(f_c_sym)
        f_llvm = ll.Function(self.module, fnty_c_func, name=name)

        self.ext_funcs[name] = f_llvm
        return {function.__qualname__: name}

    def generate(self, imports=None, system_tag=None, save_to_file=False):

        self.builder.position_at_end(self.bb_loop)

        self.builder.branch(self.bb_store)
        self.builder.position_at_end(self.bb_store)

        self.builder.branch(self.bb_exit)

        self.builder.position_at_end(self.bb_exit)

        for i, deriv in enumerate(self.derivatives):
            if deriv not in self.values:
                self._create_target_pointer(deriv)
            pointer = self._get_target_pointers([deriv])
            dg_ptr = self.builder.gep(self.deriv_global, [ll.IntType(64)(0), ll.IntType(64)(i)])
            self.builder.store(self.builder.load(*pointer), dg_ptr)

        self.builder.ret(self.builder.gep(self.deriv_global, [self.index0, self.index0]))

        # build vars read function
        self._build_var_r()

        # build vars write function
        self._build_var_w()

        if save_to_file:
            self.equations_llvm.append(str(self.module))
            for i, ll_code in enumerate(self.equations_llvm):
                llmod_saved = llvm.parse_assembly(ll_code)
                pmb = llvm.create_pass_manager_builder()
                pmb.opt_level = 3
                pm = llvm.create_module_pass_manager()
                pmb.populate(pm)
                pm.run(llmod_saved)
                self.equations_llvm_opt.append(str(llmod_saved))

        llmod = llvm.parse_assembly(str(self.module))

        pmb = llvm.create_pass_manager_builder()
        pmb.opt_level = 1
        pm = llvm.create_module_pass_manager()
        pmb.populate(pm)
        pm.run(llmod)

        ee.add_module(llmod)

        ee.finalize_object()
        cfptr = ee.get_function_address("kernel")

        cfptr_var_r = ee.get_function_address("vars_r")
        cfptr_var_w = ee.get_function_address("vars_w")

        c_float_type = type(np.ctypeslib.as_ctypes(np.float64()))

        diff_ = CFUNCTYPE(POINTER(c_float_type), POINTER(c_float_type))(cfptr)

        vars_r = CFUNCTYPE(POINTER(c_float_type), c_int64)(cfptr_var_r)

        vars_w = CFUNCTYPE(c_void_p, c_double, c_int64)(cfptr_var_w)

        n_deriv = self.n_deriv

        @njit('float64[:](float64[:])')
        def diff(y):
            deriv_pointer = diff_(y.ctypes)
            return carray(deriv_pointer, (n_deriv,)).copy()

        max_var = self.max_var

        @njit('float64[:]()')
        def read_variables():
            variables_pointer = vars_r(0)
            variables_array = carray(variables_pointer, (max_var,))

            return variables_array.copy()

        @njit('void(float64,int64)')
        def write_variables(var, idx):
            vars_w(var, idx)

        return diff, read_variables, write_variables
        # write_variables

    def detailed_print(self, *args, sep=' '):
        if config.PRINT_LLVM:
            logging.info(sep.join(map(str, args)))

    def save_module(self, filename):
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        with open(filename, 'w') as f:
            f.write(str(self.module))

    def load_global_variable(self, variable_name):
        index = ll.IntType(64)(self.variable_names[variable_name])
        ptr = self.var_global
        indices = [self.index0, index]
        eptr = self.builder.gep(ptr, indices, name="variable_" + variable_name)
        self.values[variable_name] = eptr

    def load_state_variable(self, idx, state_name):
        index_args = ll.IntType(64)(idx)
        ptr = self.func.args[0]
        indices = [index_args]
        eptr = self.builder.gep(ptr, indices, name="state_" + state_name)
        self.values[state_name] = eptr
        ptr = self.var_global
        index_global = ll.IntType(64)(self.variable_names[state_name])
        indices = [self.index0, index_global]
        eptr = self.builder.gep(ptr, indices)
        self.builder.store(self.builder.load(self.values[state_name]), eptr)

    def store_variable(self, variable_name):
        _block = self.builder.block
        self.builder.position_at_end(self.bb_store)
        index = ll.IntType(64)(self.variable_names[variable_name])
        ptr = self.var_global
        indices = [self.index0, index]
        eptr = self.builder.gep(ptr, indices)
        self.builder.store(self.builder.load(self.values[variable_name]), eptr)
        self.builder.position_at_end(_block)

    def add_call(self, external_function_name, args, target_ids):
        arg_pointers = []
        for i1, arg in enumerate(args):
            if i1 in target_ids:
                if arg not in self.values:
                    self._create_target_pointer(arg)
                arg_pointers.append(*self._get_target_pointers([arg]))
            else:
                arg_pointers.append(*self._load_args([arg]))
        self.builder.position_at_end(self.bb_loop)
        self.builder.call(self.ext_funcs[external_function_name], arg_pointers)

    def _build_var_r(self):
        fnty_vars = ll.FunctionType(ll.DoubleType().as_pointer(), [ll.IntType(64)])

        vars_func = ll.Function(self.module, fnty_vars, name="vars_r")

        bb_entry_var = vars_func.append_basic_block()

        bb_exit_var = vars_func.append_basic_block()

        builder_var = ll.IRBuilder()
        builder_var.position_at_end(bb_entry_var)
        builder_var.branch(bb_exit_var)

        builder_var.position_at_end(bb_exit_var)

        index0_var = ll.IntType(64)(0)
        vg_ptr = builder_var.gep(self.var_global, [index0_var, index0_var])
        builder_var.ret(vg_ptr)

    def _build_var_w(self):
        fnty_vars = ll.FunctionType(ll.VoidType(), [ll.DoubleType(), ll.IntType(64)])
        fnty_vars.args[0].name = "vars"
        fnty_vars.args[1].name = "idx"
        vars_func = ll.Function(self.module, fnty_vars, name="vars_w")
        bb_entry_var = vars_func.append_basic_block()
        bb_exit_var = vars_func.append_basic_block()

        builder_var = ll.IRBuilder()
        builder_var.position_at_end(bb_entry_var)
        builder_var.branch(bb_exit_var)
        builder_var.position_at_end(bb_exit_var)

        index = vars_func.args[1]
        ptr = self.var_global
        indices = [self.index0, index]
        eptr = builder_var.gep(ptr, indices)
        builder_var.store(vars_func.args[0], eptr)

        builder_var.ret_void()

    def _store_to_global_target(self, target, value):
        if target not in self.values:
            self._create_target_pointer(target)
        self.builder.store(value, self.values[target])

    def add_mapping(self, args, targets):
        self.builder.position_at_end(self.bb_loop)
        la = len(args)

        loaded_args_ptr = self._load_args(args)

        if la == 1:
            for t in targets:
                self._store_to_global_target(t, loaded_args_ptr[0])
        else:
            accum = self.builder.fadd(loaded_args_ptr[0], loaded_args_ptr[1])
            for i, a in enumerate(loaded_args_ptr[2:]):
                accum = self.builder.fadd(accum, a)

            for t in targets:
                self._store_to_global_target(t, accum)

    def _load_args(self, args):
        loaded_args = []
        self.builder.position_at_end(self.bb_loop)
        for a in args:
            if a not in self.values:
                self.load_global_variable(a)
            loaded_args.append(self.builder.load(self.values[a], 'arg_' + a))
        return loaded_args

    def _get_target_pointers(self, targets):
        return [self.values[target] for target in targets]

    def _create_target_pointer(self, t):
        index = ll.IntType(64)(self.variable_names[t])
        ptr = self.var_global
        indices = [self.index0, index]
        eptr = self.builder.gep(ptr, indices, name=t)
        self.values[t] = eptr

    ##Should be moved to utils
    def _is_range_and_ordered(self, a_list):
        e = len(a_list)
        i = a_list[0]
        return all(ele >= i and ele < i + e for ele in a_list) and all(a_list[i] <= a_list[i + 1] for i in range(e - 1))

    def add_set_call(self, external_function_name, variable_name_arg_and_trg, targets_ids):
        ## TODO check allign
        self.builder.position_at_end(self.bb_loop)
        number_of_ix_arg = len(variable_name_arg_and_trg[0])
        number_of_ix_trg = len(targets_ids)
        size = len(variable_name_arg_and_trg)
        # order_arg = [self.variable_names[x] for x in list(itertools.chain(*variable_name_arg))]
        # order_trg = [self.variable_names[x] for x in list(itertools.chain(*variable_name_trg))]
        # if not ((self._is_range_and_ordered(order_arg)) and (self._is_range_and_ordered(order_trg))):
        #     raise ValueError("llvm set call arguments are not align.")

        index_arg = (ll.IntType(64)(self.variable_names[variable_name_arg_and_trg[0][0]] - 1))
        index_arg_ptr = self.builder.alloca(ll.IntType(64))
        self.builder.store(index_arg, index_arg_ptr)

        ##crete new block

        size_ptr = ll.IntType(64)(size)

        loop_cond = self.func.append_basic_block(name='for' + str(self.loopcount) + '.cond')
        self.builder.position_at_end(loop_cond)
        i0 = self.builder.phi(ll.IntType(64), name='loop' + str(self.loopcount) + '.ix')

        self.builder.position_at_end(self.bb_loop)
        self.builder.branch(loop_cond)

        loop_body = self.func.append_basic_block(name='for' + str(self.loopcount) + '.body')
        self.builder.position_at_end(loop_body)

        loaded_args = []
        loaded_trgs = []
        for i1 in range(number_of_ix_arg):
            index_arg = self.builder.load(index_arg_ptr)
            index_arg = self.builder.add(index_arg, ll.IntType(64)(1), name="index_arg")
            self.builder.store(index_arg, index_arg_ptr)

            indices_arg = [self.index0, index_arg]
            eptr_arg = self.builder.gep(self.var_global, indices_arg, name="loop_" + str(self.loopcount) + " _arg_")
            if i1 in targets_ids:
                loaded_args.append(eptr_arg)
            else:
                loaded_args.append(self.builder.load(eptr_arg, 'arg_' + str(self.loopcount) + "_" + str(i1)))

        # EQ_system_SET_oscillators_mechanics_llvm
        self.builder.call(self.ext_funcs[external_function_name], loaded_args)

        loop_inc = self.func.append_basic_block(name='for' + str(self.loopcount) + '.inc')
        self.builder.position_at_end(loop_inc)
        result = self.builder.add(i0, ll.IntType(64)(1), name="res")
        self.builder.branch(loop_cond)

        loop_end = self.func.append_basic_block(name='main' + str(self.loopcount))

        self.builder.position_at_end(loop_cond)
        i0.add_incoming(ll.IntType(64)(0), self.bb_loop)
        i0.add_incoming(result, loop_inc)
        cmp = self.builder.icmp_unsigned("<", i0, size_ptr, name="cmp")
        self.builder.cbranch(cmp, loop_body, loop_end)

        self.builder.position_at_end(loop_body)
        self.builder.branch(loop_inc)
        self.bb_loop = loop_end

        self.loopcount += 1

    def add_set_mapping(self, args2d, targets2d):
        pass
